"""lindbladian.py

SuperOperator class for Lindbladian, used in open quantum dynamics
"""

import numpy as np
from muspinsim.celio import CelioHamiltonian

from muspinsim.hamiltonian import Hamiltonian
from muspinsim.spinop import SuperOperator, SpinOperator, DensityOperator
from muspinsim.validation import (
    validate_evolve_params,
    validate_integrate_decaying_params,
)


class Lindbladian(SuperOperator):
    @classmethod
    def from_hamiltonian(self, H, dissipators=[]):

        if isinstance(H, CelioHamiltonian):
            raise NotImplementedError(
                "Linbladian is not implemented for Celio's method"
            )
        if not isinstance(H, Hamiltonian):
            raise ValueError("Must use Hamiltonian to create Lindbladian")

        L = -1.0j * SuperOperator.commutator(H)
        L = self(L.matrix, L.dimension)

        for (A, gamma) in dissipators:
            L.add_dissipative_term(A, gamma)

        return L

    def add_dissipative_term(self, A, gamma=1.0):

        AA = A.dagger() * A
        Ld = gamma * (SuperOperator.bracket(A) - 0.5 * SuperOperator.anticommutator(AA))
        if Ld.dimension != self.dimension:
            raise ValueError("Invalid dissipation operator for this Lindbladian")

        self._matrix += Ld.matrix

    def evolve(self, rho0, times, operators=[]):
        """Time evolution of a state under this Lindbladian

        Perform an evolution of a state described by a DensityOperator under
        this Lindbladian and return either a sequence of DensityOperators or
        a sequence of expectation values for given SpinOperators.

        Arguments:
            rho0 {DensityOperator} -- Initial state
            times {ndarray} -- Times to compute the evolution for, in microseconds

        Keyword Arguments:
            operators {[SpinOperator]} -- List of SpinOperators to compute the
                                          expectation values of at each step.
                                          If omitted, the states' density
                                          matrices will be returned instead
                                           (default: {[]})

        Returns:
            [DensityOperator | ndarray] -- DensityOperators or expectation values

        Raises:
            TypeError -- Invalid operators
            ValueError -- Invalid values of times or operators
            RuntimeError -- Hamiltonian is not hermitian
        """

        times = np.array(times)
        if isinstance(operators, SpinOperator):
            operators = [operators]

        validate_evolve_params(rho0, times, operators)

        dim = rho0.dimension
        if self.dimension != dim * 2:
            raise ValueError("Incompatible rho0 dimension")

        if any([self.dimension != o.dimension * 2 for o in operators]):
            raise ValueError("Incompatible measure operator dimension")

        # Start by building the matrix
        L = self.matrix.toarray()

        # Diagonalize it
        evals, revecs = np.linalg.eig(L)

        # Vec-ing the density matrix
        rho0 = rho0.matrix.toarray().reshape(
            -1,
        )
        rho0 = np.linalg.solve(revecs, rho0)
        # And the operators
        operatorsT = np.array(
            [np.dot(o.matrix.T.toarray().reshape((-1,)), revecs) for o in operators]
        )

        rho = np.exp(2.0 * np.pi * evals[None, :] * times[:, None]) * rho0[None, :]

        if len(operators) > 0:
            # Expectation values
            result = np.sum(operatorsT[None, :, :] * rho[:, None, :], axis=-1)
        else:
            # Density matrices
            n = np.prod(dim)
            result = [
                DensityOperator(np.dot(revecs, r).reshape(n, n), dim) for r in rho
            ]

        return result

    def integrate_decaying(self, rho0, tau, operators=[]):
        """Integrate one or more expectation values in time with decay

        Perform an integral in time from 0 to +inf of an expectation value
        computed on an evolving state, times a decay factor exp(-t/tau).
        This can be done numerically using evolve, but here is executed with
        a single evaluation, making it a lot faster.

        Arguments:
            rho0 {DensityOperator} -- Initial state
            tau {float} -- Decay time, in microseconds

        Keyword Arguments:
            operators {list} -- Operators to compute the expectation values
                                of (default: {[]})

        Returns:
            ndarray -- List of integral values

        Raises:
            TypeError -- Invalid operators
            ValueError -- Invalid values of tau or operators
            RuntimeError -- Hamiltonian is not hermitian
        """

        if isinstance(operators, SpinOperator):
            operators = [operators]

        validate_integrate_decaying_params(rho0, tau, operators)

        # Start by building the matrix
        L = self.matrix

        # Diagonalize it

        # sparse matricies

        # need to convert this to use sparse matrices instead
        evals, revecs = np.linalg.eig(L.toarray())

        # evals, revecs = linalg.eigsh(self._matrix, k=self._matrix.shape[0]-2)
        # idx = evals.argsort()
        # evals = eigval[idx]
        # revecs = eigvec[:,idx]

        # Vec-ing the density matrix
        rho0 = rho0.matrix.toarray().reshape((-1,))
        rho0 = np.linalg.solve(revecs, rho0)

        # And the operators
        intops = np.array(
            [
                np.dot(o.matrix.T.toarray().reshape((-1,)), revecs)
                / (1.0 / tau - 2.0 * np.pi * evals)
                for o in operators
            ]
        )

        result = np.sum(rho0[None, :] * intops[:, :], axis=1)

        return result



"""simconfig.py

Classes to generate and distribute configurations for simulations
"""

import re
import os
import logging
import datetime
from collections import OrderedDict, namedtuple
from collections.abc import Iterable
from itertools import product

import numpy as np
from ase.quaternions import Quaternion

from muspinsim.spinsys import MuonSpinSystem
from muspinsim.utils import quat_from_polar


# A dictionary of correspondence between keyword names and config parameters
_CDICT = {
    "polarization": "mupol",
    "field": "B",
    "intrinsic_field": "intrinsic_B",
    "time": "t",
    "orientation": "orient",
    "temperature": "T",
}

# Same but specifically for coupling types
_CSDICT = {
    "zeeman": "zmn",
    "dipolar": "dip",
    "quadrupolar": "quad",
    "hyperfine": "hfc",
    "dissipation": "dsp",
}

# Named tuple for snapshots
ConfigSnapshot = namedtuple("ConfigSnapshot", ["id", "y"] + list(_CDICT.values()))


# A useful decorator
def _validate_coupling_args(fun):
    def decorated(self, value, **args):
        ij = np.array([v for v in args.values() if v is not None])
        if (ij < 1).any() or (ij > len(self.system)).any():
            raise MuSpinConfigError("Out of range indices for coupling")

        return fun(self, value, **args)

    return decorated


def _validate_shape(v, target_shape=(3,), name="vector"):
    v = np.array(v)
    if v.shape != target_shape:
        raise MuSpinConfigError("Invalid shape for {0} coupling term".format(name))
    return v


# Another utility function


def _elems_from_arrayodict(inds, odict):
    return {k: v[inds[i]] for i, (k, v) in enumerate(odict.items())}


def _log_dictranges(rdict):

    for k, r in rdict.items():
        logging.info("\t\t{k} => {n} points".format(k=k, n=len(r)))


class MuSpinConfigError(Exception):
    pass


class MuSpinConfig(object):
    """A class to store a configuration for a MuSpinSim simulation, including
    any ranges of parameters as determined by the input file."""

    def __init__(self, params={}):
        """Initialise a MuSpinConfig object

        Initialise a MuSpinConfig object from values produced by the
        .evaluate method of a MuSpinInput object.

        Arguments:
            params {dict} -- Dictionary of parameters as returned by
                             MuSpinInput.evaluate

        """

        self._constants = {}
        self._file_ranges = OrderedDict()
        self._avg_ranges = OrderedDict()
        self._x_range = OrderedDict()

        if len(params) == 0:
            return

        # Basic parameters
        self._name = self.validate("name", params["name"].value)[0]
        self._spins = self.validate("spins", params["spins"].value[0])
        self._celio = self._validate_celio(params["celio"].value[0])
        self._y_axis = self.validate("y", params["y_axis"].value[0])[0]

        # Identify ranges
        try:
            x = _CDICT[params["x_axis"].value[0][0]]
        except KeyError:
            raise MuSpinConfigError("Invalid x axis name in input file")
        self._x_range[x] = None

        for a in params["average_axes"].value.reshape((-1,)):
            if a.lower() == "none":
                # Special case
                continue
            try:
                self._avg_ranges[_CDICT[a]] = None
            except KeyError:
                raise MuSpinConfigError("Invalid average axis name in input file")

        self._time_N = 0  # Number of time points. This is special
        self._time_isavg = "t" in self._avg_ranges  # Is time averaged over?

        # Now inspect all parameters
        for iname, cname in _CDICT.items():

            try:
                p = params[iname]
            except KeyError:
                raise MuSpinConfigError(
                    "Invalid params object passed to "
                    "MuSpinConfig: "
                    "missing {0}".format(iname)
                )

            v = self.validate(cname, p.value, p.args)

            # Some additional special case treatment
            if cname == "t":
                if self._y_axis == "integral":
                    # Time is useless, might as well remove it
                    logging.warning("Ignoring time axis since Y = integral")
                    v = [np.inf]

                self._time_N = len(v)
            elif cname == "orient":
                if cname in self._avg_ranges:
                    # We need to normalize the weights so that they sum to N
                    norm = len(v) / np.sum(np.array([w for (q, w) in v]))
                    v = np.array([(q, w * norm) for (q, w) in v])
                else:
                    # Ignore the weights
                    v = np.array([(q, 1.0) for (q, w) in v])

            if len(v) > 1:
                # It's a range
                if cname in self._x_range:
                    self._x_range[cname] = v
                elif cname in self._avg_ranges:
                    self._avg_ranges[cname] = v
                else:
                    self._file_ranges[cname] = v
            else:
                # It's a constant
                self._constants[cname] = v[0]

        # Check that the Y axis and time are consistent
        if self._y_axis == "integral":
            if "t" in self._x_range:
                raise MuSpinConfigError(
                    "Can not use time as X axis when evaluating integral of signal"
                )

        # If we're fitting, we can't have file ranges
        finfo = params["fitting_info"]
        if finfo["fit"]:
            if len(self._file_ranges) > 0:
                raise MuSpinConfigError("Can not have file ranges when fitting")
            # The x axis is overridden, whatever it is
            xname = list(self._x_range.keys())[0]
            self._constants.pop(xname, None)  # Just in case it was here
            self._x_range[xname] = finfo["data"][:, 0]
            if xname == "t":
                # Special case
                self._time_N = len(self._x_range[xname])

        # Check that a X axis was found
        if list(self._x_range.values())[0] is None:
            raise MuSpinConfigError("Specified x axis is not a range")

        # Remove anything that is not a range in averages
        self._avg_ranges = OrderedDict(
            **{k: v for k, v in self._avg_ranges.items() if v is not None}
        )

        logging.info("Using X axis:")
        _log_dictranges(self._x_range)
        if len(self._avg_ranges) > 0:
            logging.info("Averaging over:")
            _log_dictranges(self._avg_ranges)
        if len(self._file_ranges) > 0:
            logging.info("Scanning over:")
            _log_dictranges(self._file_ranges)

        # Now make the spin system
        self._system = MuonSpinSystem(self._spins, self.celio_k)
        self._dissip_terms = {}

        for iid, idata in params["couplings"].items():
            iname = idata.name
            try:
                cname = _CSDICT[iname]
            except KeyError:
                raise MuSpinConfigError(
                    "Invalid params object passed to "
                    "MuSpinConfig: "
                    "unknown {0} coupling".format(iname)
                )

            cval = self.validate(cname, idata.value, idata.args)[0]

            i = idata.args.get("i")
            j = idata.args.get("j")

            # Move back from 1-based to 0-based indexing
            i = i - 1 if i is not None else None
            j = j - 1 if j is not None else None

            if cname == "zmn":
                # Zeeman coupling
                self._system.add_zeeman_term(i, cval)
            elif cname == "dip":
                # Dipolar coupling
                self._system.add_dipolar_term(i, j, cval)
            elif cname == "hfc":
                # Hyperfine coupling
                self._system.add_hyperfine_term(i, cval, j)
            elif cname == "quad":
                # Quadrupolar coupling
                self._system.add_quadrupolar_term(i, cval)
            elif cname == "dsp":
                # Dissipation. Special case, this is temperature dependent and
                # must be set individually
                self._dissip_terms[i] = cval[0]

        # Now for results, use the shapes of only file and x ranges
        res_shape = [len(v) for v in self._file_ranges.values()]
        res_shape += [len(v) for v in self._x_range.values()]

        self._results = np.zeros(res_shape)

        # And define the configurations, individually and collectively
        def make_configs(od):
            cfg = []
            for k, v in od.items():
                if k == "t":
                    cfg.append([slice(None)])
                else:
                    cfg.append(np.arange(len(v)))
            return list(product(*cfg))

        fconfigs = make_configs(self._file_ranges)
        aconfigs = make_configs(self._avg_ranges)
        xconfigs = make_configs(self._x_range)

        # Total configurations
        self._configurations = list(product(*[fconfigs, aconfigs, xconfigs]))
        # Size of the average
        self._avg_N = len(aconfigs)

        logging.info(
            "Total number of configurations to simulate: "
            "{0}".format(len(self._configurations))
        )
        logging.info(
            "Total number of configurations to average: {0}".format(self._avg_N)
        )

    def validate(self, name, value, args={}):
        """Validate an input parameter with a custom method, if present.

        Arguments:
            name {str} -- Name of the input parameter
            value {any} -- Value to validate

        Keyword arguments:
            args {dict} -- Arguments, if present

        Returns:
            value -- Validated and normalised value
        """

        vname = "_validate_{0}".format(name)

        if hasattr(self, vname):
            vfun = getattr(self, vname)
            value = [vfun(v, **args) for v in value]

        return value

    def store_time_slice(self, config_id, tslice):
        """Store a time slice of data in this configuration's results.

        Store a time slice of data, given the configuration snapshot ID, and
        sum it to pre-existing data according to averaging rules.


        Arguments:
            config_id {tuple} -- ID of the ConfigurationSnapshot with which
                                 the data was calculated
            tslice {np.ndarray} -- Time slice of data
        """

        # Check the shape
        if isinstance(tslice, Iterable) and len(tslice) != self._time_N:
            raise ValueError("Time slice has invalid length")

        if self._time_isavg:
            tslice = np.average(tslice)

        ii = tuple(list(config_id[0]) + list(config_id[2]))
        self._results[ii] += tslice / self._avg_N

    def save_output(self, name=None, path=".", extension=".dat"):
        """Save all output files for the gathered results

        Save all output files for the gathered results, using an appropriate
        output path and seed name.

        Keyword arguments:
            name {str} -- Root name to use for the files
            path {str} -- Folder path to save the files in
            extension {str} -- Extension to save the files with
        """

        from muspinsim import __version__

        if name is None:
            name = self.name

        # Header format
        file_header = """MUSPINSIM v.{version}
Output file written on {date}
Parameters used:
""".format(
            version=__version__, date=datetime.datetime.now().ctime()
        )

        for fn in self._file_ranges:
            file_header = file_header + "\t{0:<20} = {{{0}}}\n".format(fn)

        indices = [range(len(rng)) for rng in self._file_ranges.values()]
        indices = product(*indices)

        # File name
        fid_pattern = "_{}" * len(self._file_ranges)
        fname_pattern = "{name}{id}{ext}"

        x = self.x_axis_values

        if "B" in self._x_range.keys() or "intrinsic_B" in self._x_range.keys():
            x = np.linalg.norm(x, axis=-1)

        # Actually save the files
        for inds in indices:
            fid = fid_pattern.format(*inds)
            fname = fname_pattern.format(name=name, id=fid, ext=extension)
            fname = os.path.join(path, fname)

            vdict = {}
            for i, (key, val) in enumerate(self._file_ranges.items()):
                pname = "_print_{0}".format(key)
                v = val[inds[i]]
                if hasattr(self, pname):
                    v = getattr(self, pname)(v)

                vdict[key] = v

            header = file_header.format(**vdict)

            data = np.zeros((len(x), 2))
            data[:, 0] = x
            data[:, 1] = self._results[inds]

            np.savetxt(fname, data, header=header)

    @property
    def name(self):
        return self._name

    @property
    def spins(self):
        return list(self._spins)

    @property
    def celio_k(self):
        """
        Returns the value of k for Celio's method, a value of 0 indicates
        Celio's method should not be used
        """
        return self._celio[0]

    @property
    def celio_averages(self):
        """
        Returns the number of averages to use in Celio's method, a value
        of 0 indicates the random states method should not be used
        """
        return self._celio[1]

    @property
    def system(self):
        return self._system.clone()

    @property
    def constants(self):
        return {**self._constants}

    @property
    def dissipation_terms(self):
        return {**self._dissip_terms}

    @property
    def results(self):
        return np.array(self._results)

    @results.setter
    def results(self, r):
        r = np.array(r)
        if r.shape != self._results.shape:
            raise ValueError("Trying to set an invalid results array")
        self._results = r

    @property
    def x_axis(self):
        return [*self._x_range.keys()][0]

    @property
    def x_axis_values(self):
        return np.array([*self._x_range.values()][0])

    @property
    def y_axis(self):
        return self._y_axis

    def __len__(self):
        return len(self._configurations)

    def __getitem__(self, i):

        isint = type(i) == int
        if isint:
            i = slice(i, i + 1)
        elif type(i) != slice:
            raise TypeError(
                "Indices must be integer or slices, not {0}".format(type(i))
            )

        ans = []

        for (fc, ac, xc) in self._configurations[i]:

            fd = _elems_from_arrayodict(fc, self._file_ranges)
            ad = _elems_from_arrayodict(ac, self._avg_ranges)
            xd = _elems_from_arrayodict(xc, self._x_range)

            tp = ConfigSnapshot(
                id=(fc, ac, xc), y=self._y_axis, **self._constants, **fd, **ad, **xd
            )
            ans.append(tp)

        if isint:
            ans = ans[0]

        return ans

    def _validate_name(self, v):
        if len(v) > 1:
            raise MuSpinConfigError("Name must be a word without spaces")
        return v[0]

    def _validate_spins(self, v):

        # Find isotopes
        isore = re.compile("([0-9]+)([A-Z][a-z]*|e)")
        m = isore.match(v)
        if m is None:
            return v
        else:
            A, el = m.groups()
            A = int(A)
            return (el, A)

    def _validate_celio(self, v):
        trotter_k = 0
        averages = 0
        # When no average is given assume only value is that of the value of k
        try:
            trotter_k = int(v[0])
            if trotter_k < 0:
                raise MuSpinConfigError(
                    "Value of k for Celio's method must a postive integer or 0"
                )
        except ValueError:
            raise MuSpinConfigError(
                "Value of k for Celio's method must be an integer"
            ) from ValueError
        # When there are 2 values, assign the value of averages as well
        if len(v) == 2:
            try:
                averages = int(v[1])
                if averages < 0:
                    raise MuSpinConfigError(
                        "Value of averages for Celio's method must a postive "
                        "integer or 0"
                    )
            except ValueError:
                raise MuSpinConfigError(
                    "Value of averages for Celio's method must be an integer"
                ) from ValueError
        return (trotter_k, averages)

    def _validate_t(self, v):
        if len(v) != 1:
            raise MuSpinConfigError("Invalid line in time range")
        return v[0]

    def _validate_B(self, v):

        if len(v) == 1:
            v = np.array([0, 0, v[0]])  # The default direction is Z
        elif len(v) != 3:
            raise MuSpinConfigError("Invalid magnetic field value")

        return v

    def _validate_intrinsic_B(self, v):

        if len(v) == 1:
            v = np.array([0, 0, v[0]])  # The default direction is Z
        elif len(v) != 3:
            raise MuSpinConfigError("Invalid intrinsic magnetic field value")

        return v

    def _validate_mupol(self, v):
        v = np.array(v, dtype=float)
        if len(v) != 3:
            raise MuSpinConfigError("Invalid muon polarization direction")

        v /= np.linalg.norm(v)

        return v

    def _validate_orient(self, v, mode):

        q = None
        w = 1.0  # Weight
        if len(v) == 2:
            # Interpret them as polar angles
            q = quat_from_polar(*v)
        elif len(v) == 3:
            q = Quaternion.from_euler_angles(*v, mode=mode)
        elif len(v) == 4:
            q = Quaternion.from_euler_angles(*v[:3], mode=mode)
            w = v[3]

        # After computing the rotation, we store the conjugate because it's a
        # lot cheaper, instead of rotating the whole system by q, to rotate
        # only the external magnetic field and the polarization (lab frame) by
        # the inverse of q

        return (q.conjugate(), w)

    def _validate_T(self, v):
        return v[0]

    @_validate_coupling_args
    def _validate_zmn(self, v, i):
        return _validate_shape(v, (3,), "Zeeman")

    @_validate_coupling_args
    def _validate_dip(self, v, i, j):
        return _validate_shape(v, (3,), "dipolar")

    @_validate_coupling_args
    def _validate_hfc(self, v, i, j=None):
        return _validate_shape(v, (3, 3), "hyperfine")

    @_validate_coupling_args
    def _validate_quad(self, v, i):
        return _validate_shape(v, (3, 3), "quadrupolar")

    @_validate_coupling_args
    def _validate_dsp(self, v, i):
        return _validate_shape(v, (1,), "dissipation")

    def _print_B(self, v):
        return "{0} T".format(v)

    def _print_orient(self, v):
        o, w = v
        abc = o.euler_angles("zyz") * 180 / np.pi
        ostr = (
            "[ZYZ] a = {0:.1f} deg, b = {1:.1f} deg, c = {2:.1f} deg, weight = {3}"
        ).format(*abc, w)
        return ostr



import unittest
import numpy as np
from io import StringIO
from ase.quaternions import Quaternion

from muspinsim.input import MuSpinInput
from muspinsim.input.input import MuSpinInputError
from muspinsim.simconfig import MuSpinConfig, MuSpinConfigError


class TestConfig(unittest.TestCase):
    def test_config(self):

        stest = StringIO(
            """
spins
    mu 2H e
field
    range(1, 11, 2)
intrinsic_field
    range(1, 21, 2)
temperature
    range(0, 10, 2)
time
    range(0, 10, 21)
orientation
    0 0
    0 180
polarization
    0 1 1
zeeman 1
    1 0 0
dipolar 1 2
    0 1 0
hyperfine 1 3
    10 2 2
    2 10 2
    2 2 10
quadrupolar 2
    -2 0 0
    0  1 0
    0  0 1
dissipation 2
    0.1
"""
        )

        itest = MuSpinInput(stest)

        cfg = MuSpinConfig(itest.evaluate())

        self.assertEqual(cfg.name, "muspinsim")
        self.assertEqual(len(cfg), 16)
        self.assertEqual(cfg.results.shape, (2, 2, 2, 21))

        # Try getting one configuration snapshot
        cfg0 = cfg[0]

        self.assertTrue((cfg0.B == [0, 0, 1]).all())
        self.assertTrue((cfg0.intrinsic_B == [0, 0, 1]).all())
        self.assertEqual(cfg0.T, 0)
        self.assertTrue((cfg0.t == np.linspace(0, 10, 21)).all())
        self.assertTrue(np.isclose(np.linalg.norm(cfg0.mupol), 1.0))

        # Now try recording results
        for c in cfg:
            res = np.ones(len(c.t))
            cfg.store_time_slice(c.id, res)
        self.assertTrue((cfg.results == 1).all())

        # System checks
        self.assertEqual(cfg.system.spins[0], "mu")
        self.assertEqual(cfg.system.spins[1], ("H", 2))
        self.assertEqual(cfg.system.spins[2], "e")
        self.assertEqual(len(cfg.system._terms), 4)
        self.assertEqual(cfg._dissip_terms[1], 0.1)

        self.assertIn("B", cfg._file_ranges)
        self.assertIn("intrinsic_B", cfg._file_ranges)
        self.assertIn("T", cfg._file_ranges)
        self.assertIn("t", cfg._x_range)
        self.assertIn("orient", cfg._avg_ranges)

        # Now try a few errors

        stest = StringIO(
            """
spins
    mu e
zeeman 3
    0 0 1
"""
        )

        itest = MuSpinInput(stest)

        with self.assertRaises(MuSpinConfigError):
            cfg = MuSpinConfig(itest.evaluate())

        stest = StringIO(
            """
spins
    mu e
zeeman 1
    2 0
"""
        )
        with self.assertRaises(MuSpinInputError):
            itest = MuSpinInput(stest)
            cfg = MuSpinConfig(itest.evaluate())

        stest = StringIO(
            """
y_axis
    integral
"""
        )

        itest = MuSpinInput(stest)

        with self.assertRaises(MuSpinConfigError):
            cfg = MuSpinConfig(itest.evaluate())

    def test_celio(self):
        # No mention of celio
        stest = StringIO(
            """
spins
    mu e
"""
        )

        itest = MuSpinInput(stest)
        cfg = MuSpinConfig(itest.evaluate())
        self.assertEqual(cfg.celio_k, 0)
        self.assertEqual(cfg.celio_averages, 0)

        # Only specifying a k
        stest = StringIO(
            """
spins
    mu e
celio
    10
"""
        )

        itest = MuSpinInput(stest)
        cfg = MuSpinConfig(itest.evaluate())
        self.assertEqual(cfg.celio_k, 10)
        self.assertEqual(cfg.celio_averages, 0)

        # Using default k and averages
        stest = StringIO(
            """
spins
    mu e
celio
    0 0
"""
        )

        itest = MuSpinInput(stest)
        cfg = MuSpinConfig(itest.evaluate())
        self.assertEqual(cfg.celio_k, 0)
        self.assertEqual(cfg.celio_averages, 0)

        # k and averages
        stest = StringIO(
            """
spins
    mu e
celio
    10 8
"""
        )

        itest = MuSpinInput(stest)
        cfg = MuSpinConfig(itest.evaluate())
        self.assertEqual(cfg.celio_k, 10)
        self.assertEqual(cfg.celio_averages, 8)

        # Try a few errors
        stest = StringIO(
            """
spins
    mu e
celio
    10 -8
"""
        )

        itest = MuSpinInput(stest)
        with self.assertRaises(MuSpinConfigError):
            cfg = MuSpinConfig(itest.evaluate())

        stest = StringIO(
            """
spins
    mu e
celio
    -1 0
"""
        )

        itest = MuSpinInput(stest)
        with self.assertRaises(MuSpinConfigError):
            cfg = MuSpinConfig(itest.evaluate())

    def test_orient(self):
        # Some special tests to check how orientations are dealt with

        stest = StringIO(
            """
orientation
    0  0  0  1.0
    0.5*pi 0 0 2.0
"""
        )

        itest = MuSpinInput(stest)

        cfg = MuSpinConfig(itest.evaluate())

        orange = cfg._avg_ranges["orient"]

        self.assertEqual(sum([w for (q, w) in orange]), 2.0)
        self.assertTrue(
            np.isclose(orange[1][0].q, [2 ** (-0.5), 0, 0, -(2 ** (-0.5))]).all()
        )

        # Some more complex euler angles combinations
        rng = np.linspace(0, np.pi, 4)
        angles = np.array(np.meshgrid(*[rng, rng, rng])).reshape((3, -1)).T
        ablock = "\n".join(map(lambda x: "\t{0} {1} {2}".format(*x), angles))

        stest = StringIO("orientation\n" + ablock)

        itest = MuSpinInput(stest)

        cfg = MuSpinConfig(itest.evaluate())

        for ((a, b, c), (q1, w)) in zip(angles, cfg._avg_ranges["orient"]):
            q2 = Quaternion.from_axis_angle([0, 0, 1], c)
            q2 *= Quaternion.from_axis_angle([0, 1, 0], b)
            q2 *= Quaternion.from_axis_angle([0, 0, 1], a)
            q2 = q2.conjugate()

            self.assertTrue(np.isclose(q1.q, q2.q).all())

        # Same, but for mode zxz
        stest = StringIO("orientation zxz\n" + ablock)

        itest = MuSpinInput(stest)

        cfg = MuSpinConfig(itest.evaluate())

        for ((a, b, c), (q1, w)) in zip(angles, cfg._avg_ranges["orient"]):
            q2 = Quaternion.from_axis_angle([0, 0, 1], c)
            q2 *= Quaternion.from_axis_angle([1, 0, 0], b)
            q2 *= Quaternion.from_axis_angle([0, 0, 1], a)
            q2 = q2.conjugate()

            self.assertTrue(np.isclose(q1.q, q2.q).all())

        # Same, but for theta and phi angles alone
        angles = np.array(np.meshgrid(*[rng, rng])).reshape((2, -1)).T
        ablock = "\n".join(map(lambda x: "\t{0} {1}".format(*x), angles))

        stest = StringIO("orientation\n" + ablock)

        itest = MuSpinInput(stest)

        cfg = MuSpinConfig(itest.evaluate())

        for ((theta, phi), (q1, w)) in zip(angles, cfg._avg_ranges["orient"]):
            q2 = Quaternion.from_axis_angle([0, 0, 1], phi)
            q2 *= Quaternion.from_axis_angle([0, 1, 0], theta)
            q2 *= Quaternion.from_axis_angle([0, 0, 1], phi)
            q2 = q2.conjugate()

            self.assertTrue(np.isclose(q1.q, q2.q).all())

    def test_fitting(self):
        # Special tests for fitting tasks

        stest = StringIO(
            """
fitting_variables
    x
fitting_data
    0   0.5
    1   0.2
"""
        )

        itest = MuSpinInput(stest)
        cfg = MuSpinConfig(itest.evaluate(x=0.0))
        # Check that the time axis has been overridden
        self.assertTrue((np.array(cfg._x_range["t"]) == [0, 1]).all())

        # Should fail due to file_ranges
        stest = StringIO(
            """
fitting_variables
    x
fitting_data
    0   0.5
    1   0.2
average_axes
    none
orientation
    0  0
    0  1
"""
        )

        itest = MuSpinInput(stest)
        with self.assertRaises(MuSpinConfigError):
            cfg = MuSpinConfig(itest.evaluate(x=0.0))



import numpy as np
from cython.parallel import prange

cimport numpy as np
cimport cython

np.import_array()

# Use C versions to avoid linking back to python
cdef extern from "<math.h>" nogil:
    double cos(double theta)
    double sin(double theta)

@cython.boundscheck(False) # Disable bounds-checking
@cython.wraparound(False)  # Disable negative index wrapping
def parallel_fast_time_evolve(double [:] times, int other_dimension, double [:, :] A, double [:, :] W):
    """Computes the result of time evolution of a muon polarisation

    Fast method for computing the result of time evolution of a
    muon polarisation. To be accurate it requires that
    \frac{e \hbar^2 B}{2 m_p k T} -> 0 and so is most suitable when T is
    \inf or B is 0. 

    Calculated as:
    \frac{1}{2d}\sum_{\alpha, \beta}|<\alpha|\sigma_{\mu}^{\hat{n}}|\beta>|^2
    +
    \frac{1}{d}\sum_{\alpha < \beta}|<\alpha|\sigma_{\mu}^{\hat{n}}|\beta>|^2
        \cos[(E_{\alpha} - E{\alpha})t]

    Arguments:
        times {ndarray} -- Times to compute the evolution for, in microseconds
        other_dimension {int} -- Value of the dimension labelled as d in the
                                 equation above. Equal to half the total
                                 dimension of the system.
        A {ndarray} -- Matrix indexed by \alpha and \beta that
                       gives the amplitude
                       |<\alpha|\sigma_{\mu}^{\hat{n}}|\beta>|^2
                       of each cosine term in the sum
        W {ndarray} -- Matrix indexed by \alpha and \beta that contains the
                       differences of their eigenvalues, expressed as angular
                       frequency and computed with np.outer

    Returns:
        [ndarray] -- Expectation values
    """

    cdef Py_ssize_t num_times = times.shape[0]
    cdef Py_ssize_t mat_dim = A.shape[0]

    # Use a memory view to the results array to make access faster
    cdef np.ndarray[np.float64_t, ndim=1] results = np.zeros((num_times), dtype=np.float64)
    cdef double [:] results_view = results
    
    cdef Py_ssize_t i, j, k

    cdef double one_over_d = 1.0 / other_dimension

    # Run times in parallel
    for i in prange(num_times, nogil=True):
        for j in range(mat_dim):
            # Term where j = k
            # Multiply by 0.5 here as want 1/2d here, and 1/d below
            results_view[i] += 0.5 * A[j, j]
            # Terms where k < j
            for k in range(0, j):
                results_view[i] += A[j, k] * cos(W[j, k] * times[i])
        results_view[i] *= one_over_d

    return results


"""
MuSpinSim

A software to simulate the quantum dynamics of muon spin systems

Author: Simone Sturniolo

Copyright 2022 Science and Technology Facilities Council
This software is distributed under the terms of the MIT License
Please refer to the file LICENSE for the text of the license

"""

import os
import logging
import argparse as ap
from datetime import datetime

from muspinsim.mpi import mpi_controller as mpi
from muspinsim.input import MuSpinInput
from muspinsim.experiment import ExperimentRunner
from muspinsim.fitting import FittingRunner

LOGFORMAT = "[%(levelname)s] [%(threadName)s] [%(asctime)s] %(message)s"


def ensure_dir_path_exists(path_string):
    if not os.path.isdir(path_string):
        try:
            os.makedirs(path_string)
        except OSError:
            raise NotADirectoryError(path_string)
    return path_string


def main(use_mpi=False):
    if use_mpi:
        mpi.connect()

    if mpi.is_root:
        # Entry point for script
        parser = ap.ArgumentParser(
            description="Muspinsim - A program designed to carry out "
            "spin dynamics calculations for muon science experiments"
        )
        parser.add_argument(
            "-f",
            "--fit-report-path",
            type=str,
            default=None,
            help="filepath to store fit report if fitting parameters given",
        )
        parser.add_argument(
            "-l",
            "--log-path",
            type=str,
            default=None,
            help="filepath to store simulation logs",
        )
        parser.add_argument(
            "-o",
            "--out-dir",
            type=str,
            default=None,
            help="folder to store the output .dat files",
        )
        parser.add_argument(
            "input_file",
            type=ap.FileType("r"),
            default=None,
            help="path to file specifying simulation input parameters. "
            "For formatting and keywords, see https://"
            "muon-spectroscopy-computational-project.github.io"
            "/muspinsim/input/.",
        )
        args = parser.parse_args()
        inp_filepath = args.input_file.name

        # get input file name and directory - to be used as defaults
        inp_file_name = os.path.basename(inp_filepath).split(".")[0]
        inp_dir = os.path.dirname(inp_filepath)

        fs = open(inp_filepath)
        infile = MuSpinInput(fs)
        is_fitting = len(infile.variables) > 0

        # Open logfile
        logfile = os.path.join(inp_dir, inp_file_name + ".log")
        if args.log_path:
            try:
                # check if directory exists, if not create it
                log_path = ensure_dir_path_exists(os.path.dirname(args.log_path))
                log_fname = os.path.basename(args.log_path)
            except NotADirectoryError:
                # if log_path was just a filename use inp_dir as a default
                log_path = inp_dir
                log_fname = args.log_path
            logfile = os.path.join(log_path, log_fname)

        logging.basicConfig(
            filename=logfile,
            filemode="w",
            level=logging.INFO,
            format=LOGFORMAT,
            datefmt="%Y-%m-%d %H:%M:%S",
            force=True,
        )

        logging.info(
            "Launching MuSpinSim calculation from file: {0}".format(inp_filepath)
        )

        if is_fitting:
            logging.info(
                "Performing fitting in variables: "
                "{0}".format(", ".join(infile.variables))
            )

        tstart = datetime.now()
    else:
        infile = MuSpinInput()
        is_fitting = False

    is_fitting = mpi.broadcast(is_fitting)

    out_path = inp_dir
    if args.out_dir:
        out_path = ensure_dir_path_exists(args.out_dir)

    if not is_fitting:
        # No fitting
        runner = ExperimentRunner(infile, {})
        runner.run()

        if mpi.is_root:
            # Output
            runner.config.save_output(name=None, path=out_path)
    else:
        fitter = FittingRunner(infile)
        fitter.run(name=None, path=out_path)

        if mpi.is_root:
            rep_dname = inp_dir
            rep_fname = "{0}_fit_report.txt".format(inp_file_name)
            if args.fit_report_path:
                try:
                    # check if directory exists, if not create it
                    ensure_dir_path_exists(os.path.dirname(args.fit_report_path))
                    rep_fname = os.path.basename(args.fit_report_path)
                    rep_dname = os.path.dirname(args.fit_report_path)
                except NotADirectoryError:
                    # if fit_report_path was just a filename use inp_dir as a default
                    rep_fname = args.fit_report_path
                    rep_dname = inp_dir

            fitter.write_report(fname=rep_fname, path=rep_dname)

    if mpi.is_root:
        tend = datetime.now()
        simtime = (tend - tstart).total_seconds()
        logging.info("Simulation completed in {0:.3f} seconds".format(simtime))

    logging.shutdown()


def main_mpi():
    main(use_mpi=True)


if __name__ == "__main__":
    main()



import unittest
import numpy as np
from io import StringIO

from muspinsim.input import MuSpinInput
from muspinsim.fitting import FittingRunner


class TestFitting(unittest.TestCase):
    def test_fitset(self):

        s1 = StringIO(
            """
fitting_variables
    x  0.0
    y  1.0 0 2.0
fitting_data
    0   0.5
    1   0.5
    2   0.5
"""
        )

        i1 = MuSpinInput(s1)
        f1 = FittingRunner(i1)

        self.assertEqual(f1._xnames, ("x", "y"))
        self.assertTrue((f1._x == [0, 1]).all())
        self.assertEqual(f1._xbounds[0], (-np.inf, np.inf))
        self.assertEqual(f1._xbounds[1], (0.0, 2.0))

    def test_fitrun(self):

        # Try fitting a very basic exponential decay
        data = np.zeros((100, 3))
        data[:, 0] = np.linspace(0, 10.0, len(data))
        g = 0.2
        data[:, 1] = 0.5 * np.exp(-g * data[:, 0])
        dblock = "\n".join(["\t{0} {1}".format(*d) for d in data])

        s1 = StringIO(
            """
spins
    mu
fitting_variables
    g   0.5
fitting_data
{data}
dissipation 1
    g
""".format(
                data=dblock
            )
        )

        i1 = MuSpinInput(s1)
        f1 = FittingRunner(i1)

        sol = f1.run()

        self.assertAlmostEqual(sol.x[0], g, 3)



import sys
import os
import unittest
import tempfile
import logging

from muspinsim.__main__ import main


def set_sys_argv(vals):
    sys.argv[1:] = vals


def run_experiment(inp_string, inp_file_path, cmd_args):
    with open(inp_file_path, "w") as f:
        f.write(inp_string)
    set_sys_argv([inp_file_path] + cmd_args)
    main()


class TestCommandLineArgs(unittest.TestCase):
    @classmethod
    def tearDownClass(self):
        # The tests in this class setup the logger to use a temporary folder but for
        # other tests that run afterwards this may no longer exist so revert to the
        # default again and print to stdout to avoid FileNotFound errors
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG, force=True)

    def test_main_default_args(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            cmd_args = []
            inp_file_path = os.path.join(tmp_dir, "custom_args.in")
            run_experiment(
                """
name
    test_1
spins
    mu H
zeeman 1
    1 0 0
""",
                inp_file_path,
                cmd_args,
            )

            self.assertTrue(os.path.exists(os.path.join(tmp_dir, "custom_args.log")))
            self.assertTrue(os.path.exists("{0}/test_1.dat".format(tmp_dir)))

    def test_main_custom_args(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            with tempfile.TemporaryDirectory() as out_tmp_dir:
                cmd_args = [
                    "-o",
                    f"{out_tmp_dir}",
                    "-l",
                    f"{out_tmp_dir}/new_test.log",
                ]
                inp_file_path = os.path.join(tmp_dir, "custom_args.in")
                run_experiment(
                    """
name
    test_2
spins
    mu H
zeeman 1
    1 0 0
""",
                    inp_file_path,
                    cmd_args,
                )

                self.assertTrue(
                    os.path.exists(os.path.join(out_tmp_dir, "new_test.log"))
                )
                self.assertTrue(os.path.exists(os.path.join(out_tmp_dir, "test_2.dat")))

    def test_main_fitting_default_args(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            cmd_args = []
            inp_file_path = os.path.join(tmp_dir, "fitting_default_args.in")
            run_experiment(
                """
name
    test_fitting
spins
    mu
field
    1.0/muon_gyr
fitting_data
    0 1
    1 2
fitting_variables
    g 0.1 0.0 inf
dissipation 1
    g
""",
                inp_file_path,
                cmd_args,
            )

            self.assertTrue(
                os.path.exists(os.path.join(tmp_dir, "fitting_default_args.log"))
            )
            self.assertTrue(os.path.exists(os.path.join(tmp_dir, "test_fitting.dat")))
            self.assertTrue(
                os.path.exists(
                    os.path.join(tmp_dir, "fitting_default_args_fit_report.txt")
                )
            )

    def test_main_filenames_as_input(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            cmd_args = ["-l", "new_test.log", "-f", "new_fit_report.txt"]
            inp_file_path = os.path.join(tmp_dir, "fitting_default_args.in")
            run_experiment(
                """
name
    test_fitting
spins
    mu
field
    1.0/muon_gyr
fitting_data
    0 1
    1 2
fitting_variables
    g 0.1 0.0 inf
dissipation 1
    g
""",
                inp_file_path,
                cmd_args,
            )

            self.assertTrue(os.path.exists("{0}/new_test.log".format(tmp_dir)))
            self.assertTrue(os.path.exists("{0}/test_fitting.dat".format(tmp_dir)))
            self.assertTrue(os.path.exists("{0}/new_fit_report.txt".format(tmp_dir)))

    def test_main_fitting_custom_args(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            with tempfile.TemporaryDirectory() as out_tmp_dir:
                cmd_args = [
                    "-o",
                    f"{out_tmp_dir}",
                    "-l",
                    f"{out_tmp_dir}/new_test.log",
                    "-f",
                    f"{out_tmp_dir}/new_fit_report.txt",
                ]
                inp_file_path = os.path.join(tmp_dir, "test.in")
                run_experiment(
                    """
name
    test_fitting
spins
    mu
field
    1.0/muon_gyr
fitting_data
    0 1
    1 2
fitting_variables
    g 0.1 0.0 inf
dissipation 1
    g
""",
                    inp_file_path,
                    cmd_args,
                )

                self.assertTrue(
                    os.path.exists(os.path.join(out_tmp_dir, "new_test.log"))
                )
                self.assertTrue(
                    os.path.exists(os.path.join(out_tmp_dir, "test_fitting.dat"))
                )
                self.assertTrue(
                    os.path.exists(os.path.join(out_tmp_dir, "new_fit_report.txt"))
                )



"""keyword.py

Classes to define and read conveniently individual keywords of an input file
"""

import re
import sys
import inspect
import numpy as np

from muspinsim.constants import MU_GAMMA
from muspinsim.input.larkeval import LarkExpression, lark_tokenize
from muspinsim.input.variables import FittingVariable
from muspinsim.utils import deepmap, zcw_gen, eulrange_gen

# Supported math functions
_math_functions = {
    "sin": np.sin,
    "cos": np.cos,
    "tan": np.tan,
    "arcsin": np.arcsin,
    "arccos": np.arccos,
    "arctan": np.arctan,
    "arctan2": np.arctan2,
    "exp": np.exp,
    "log": np.log,
    "sqrt": np.sqrt,
}

# And math constants
_math_constants = {"pi": np.pi, "deg": np.pi / 180.0, "e": np.exp(1), "inf": np.inf}

_phys_constants = {"muon_gyr": MU_GAMMA, "MHz": 1.0 / (2 * MU_GAMMA)}

# Functions for powder orientation

_pwd_functions = {"zcw": zcw_gen, "eulrange": eulrange_gen}


# Expansion functions
def _range(x1, x2, n=100):
    return np.linspace(x1, x2, int(n))[:, None]


class MuSpinKeyword(object):
    """Generic class used to parse a keyword from a MuSpinSim input file"""

    name = "keyword"
    block_size = 1
    block_size_bounds = (1, np.inf)
    accept_range = True
    accept_as_x = False
    default = None
    expr_size_bounds = (1, np.inf)
    _validators = []

    def __init__(self, block=[], args=[]):
        """Create an instance of a given keyword, passing the raw block of
        text as well as the arguments.

        Arguments:
            block {[str]} -- Lines of text defining the value of the keyword
            args {[any]} -- Any arguments appearing after the keyword

        """

        # Sanity check
        if (
            self.block_size < self.block_size_bounds[0]
            or self.block_size > self.block_size_bounds[1]
        ):
            raise RuntimeError("Invalid block_size for keyword {0}".format(self.name))

        self._store_args(args)

        # Reshape block
        block = np.array(block)
        try:
            block = block.reshape((-1, self.block_size))
        except ValueError:
            raise RuntimeError(
                "Invalid number of entries for block, expected {0}, got {1}".format(
                    self.block_size, len(block)
                )
            )
        if not self.accept_range and len(block) > 1:
            raise RuntimeError(
                "Can not accept range of values for keyword {0}".format(self.name)
            )

        use_default = False
        if len(block) == 0 or (
            len(block) == 1
            and not type(block[0] is np.ndarray)
            and block[0] in ("", None)
        ):
            if self.has_default:
                block = np.array([self.default.split("\n")])
                use_default = True
            else:
                raise RuntimeError(
                    "Input is empty and keyword '{0}' "
                    "doesn't have a default value".format(self.name)
                )

        self._store_values(block)

        self._validate_values()

        # check each line of block to make sure number of
        # expressions are within bounds (if not using defaults)
        # need to check after parsing to account for expressions and functions
        if not use_default:
            if type(self._values) is np.ndarray:
                entry_lengths = [
                    np.shape(self._values)[1] for _ in range(np.shape(self._values)[0])
                ]
            else:
                # traverse nested lists
                nested_val = self._values

                while isinstance(nested_val[0], list) and len(nested_val) == 1:
                    nested_val = nested_val[0]

                entry_lengths = (
                    [len(entry) for entry in nested_val]
                    if isinstance(nested_val[0], list)
                    else [len(nested_val)]
                )

            for i, length in enumerate(entry_lengths):
                if (
                    length < self.expr_size_bounds[0]
                    or length > self.expr_size_bounds[1]
                ):
                    raise RuntimeError(
                        "Incorrect number of args for entry '{0}', "
                        "expected {1}, got {2}".format(
                            block[i][0],
                            "between {0} and {1}".format(
                                self.expr_size_bounds[0],
                                self.expr_size_bounds[1],
                            )
                            if self.expr_size_bounds[0] != self.expr_size_bounds[1]
                            else self.expr_size_bounds[0],
                            length,
                        )
                    )

    def _default_args(self):
        # Dummy function, used for type signature and processing of arguments
        return {}

    def _store_args(self, args):
        try:
            self._args = self._default_args(*args)
        except TypeError:
            if args:
                raise RuntimeError(
                    "Wrong number of in-line arguments given '{0}', "
                    "expected {1}, got {2}".format(
                        " ".join(args),
                        len(inspect.signature(self._default_args).parameters),
                        len(args),
                    )
                )
            else:
                raise RuntimeError(
                    "This keyword requires {0} in-line arguments".format(
                        len(inspect.signature(self._default_args).parameters),
                    )
                )
        except ValueError as e:
            raise RuntimeError(
                "Error parsing keyword argument(s) '{0}': {1}".format(self.name, str(e))
            )

    def _store_values(self, block):
        # Parse and store each value separately
        self._values = []
        for v in block:
            b = [l.split() for l in v]
            if len(b) == 1:
                b = b[0]
            self._values.append(b)
        self._values = np.array(self._values)

    def _validate_values(self):
        errs = ""
        for vfunc in self._validators:
            errs += "\n".join([vfunc(b) for b in self._values if vfunc(b) != ""])

        if errs:
            raise ValueError(errs)

    @property
    def arguments(self):
        return {**self._args}

    @property
    def id(self):
        return self.name

    @property
    def has_default(self):
        return self.default is not None

    def evaluate(self):
        return self._values.copy()

    def __len__(self):
        return len(self._values)


class MuSpinEvaluateKeyword(MuSpinKeyword):
    """Specialised class for keywords with values that need evaluating"""

    name = "evaluate_keyword"
    _functions = {**_math_functions}
    _constants = {**_math_constants}

    def __init__(self, block=[], args=[], variables=[]):

        cnames = set(self._constants.keys())
        fnames = set(self._functions.keys())
        vnames = set(variables)

        conflicts = cnames.intersection(vnames)
        conflicts = conflicts.union(fnames.intersection(vnames))
        if len(conflicts) > 0:
            raise ValueError(
                "Variable names {0}".format(conflicts)
                + " conflict with existing constants"
            )

        self._variables = list(cnames.union(vnames))
        super(MuSpinEvaluateKeyword, self).__init__(block, args)

    def _store_values(self, block):
        self._values = []
        for v in block:
            b = [
                [
                    LarkExpression(
                        tk, variables=self._variables, functions=self._functions
                    )
                    for tk in lark_tokenize(l)
                ]
                for l in v
            ]
            if len(b) == 1:
                b = b[0]
            self._values.append(b)

    def evaluate(self, **variables):

        allvars = {**variables, **self._constants}

        def expreval(expr):
            return expr.evaluate(**allvars)

        return np.array(deepmap(expreval, self._values))


class MuSpinExpandKeyword(MuSpinEvaluateKeyword):
    """Specialised class for keywords whose lines can be expanded with
    particular functions"""

    name = "expand_keyword"
    block_size_bounds = (1, 1)
    _functions = {**_math_functions, "range": _range}

    def evaluate(self, **variables):

        allvars = {**variables, **self._constants}

        if self.block_size != 1:
            raise RuntimeError("MuSpinExpandKeyword can not have block_size > 1")

        def expreval(expr):
            return expr.evaluate(**allvars)

        eval_values = []
        for line in self._values:
            eval_line = np.array([expreval(expr) for expr in line])

            if len(line) == 1 and len(eval_line.shape) == 3:
                eval_values += list(eval_line[0])
            elif len(line) == 1 and len(eval_line.shape) == 2:
                eval_values += [eval_line[0]]
            elif len(eval_line.shape) == 1:
                eval_values += [eval_line]
            else:
                raise RuntimeError(
                    "Unable to evaluate expression for keyword {0}".format(self.name)
                )

        return eval_values


class MuSpinCouplingKeyword(MuSpinEvaluateKeyword):

    name = "coupling_keyword"
    block_size = 1
    default = "0 0 0"

    def _default_args(self, i=None, j=None):
        args = {
            "i": int(i) if (i is not None) else None,
            "j": int(j) if (j is not None) else None,
        }
        return args

    @property
    def accept_range(self):
        return False  # Never ranges! These keywords define the system!

    @property
    def id(self):
        id_str = ""
        i = self._args.get("i")
        j = self._args.get("j")
        if i is not None:
            id_str += "_{0}".format(i)
            if j is not None:
                id_str += "_{0}".format(j)
        return "{0}{1}".format(self.name, id_str)


# Now on to defining the actual keywords that are admitted in input files
class KWName(MuSpinKeyword):

    name = "name"
    block_size = 1
    accept_range = False
    default = "muspinsim"


class KWSpins(MuSpinKeyword):

    name = "spins"
    expr_size_bounds = (1, np.inf)
    block_size = 1
    accept_range = False
    default = "mu e"


class KWCelio(MuSpinKeyword):

    name = "celio"
    expr_size_bounds = (1, 2)
    block_size = 1
    accept_range = False
    default = "0 0"


class KWPolarization(MuSpinExpandKeyword):

    name = "polarization"
    expr_size_bounds = (1, 3)
    block_size = 1
    default = "transverse"
    _constants = {
        **_math_constants,
        "longitudinal": np.array([0, 0, 1.0]),
        "transverse": np.array([1.0, 0, 0]),
    }
    _functions = {**_math_functions}


class KWField(MuSpinExpandKeyword):

    name = "field"
    block_size = 1
    accept_range = True
    accept_as_x = True
    default = "0.0"
    _constants = {**_math_constants, **_phys_constants}


class KWIntrinsicField(MuSpinExpandKeyword):

    name = "intrinsic_field"
    block_size = 1
    accept_range = True
    accept_as_x = True
    default = "0.0"
    _constants = {**_math_constants, **_phys_constants}


class KWTime(MuSpinExpandKeyword):

    name = "time"
    block_size = 1
    accept_range = True
    accept_as_x = True
    default = "range(0, 10, 101)"


class KWXAxis(MuSpinKeyword):

    name = "x_axis"
    block_size = 1
    accept_range = False
    default = "time"
    _validators = [
        lambda s: "Invalid value {0}, accepts {1}".format(
            s[0], [i for i in InputKeywords if InputKeywords[i].accept_as_x]
        )
        if s[0] not in InputKeywords or not InputKeywords[s[0]].accept_as_x
        else ""
    ]


class KWYAxis(MuSpinKeyword):

    name = "y_axis"
    block_size = 1
    accept_range = False
    default = "asymmetry"
    _validators = [
        lambda s: "Invalid value '{0}', accepts ['asymmetry', 'integral']".format(s)
        if s not in ["asymmetry", "integral"]
        else ""
    ]


class KWAverageAxes(MuSpinKeyword):

    name = "average_axes"
    block_size = 1
    accept_range = True
    default = "orientation"
    _validators = [
        lambda s: "Invalid value(s) '{0}': accepts {1}".format(
            [w for w in s if (w not in InputKeywords) or w.lower() == "None"],
            InputKeywords,
        )
        if not all((w in InputKeywords or w.lower() == "none") for w in s)
        else ""
    ]


class KWOrientation(MuSpinExpandKeyword):

    name = "orientation"
    expr_size_bounds = (1, 4)
    block_size = 1
    accept_range = True
    default = "0 0 0"
    _functions = {**_math_functions, **_pwd_functions}

    def _default_args(self, mode="zyz"):
        args = {"mode": mode}
        return args


class KWTemperature(MuSpinExpandKeyword):

    name = "temperature"
    block_size = 1
    accept_range = True
    accept_as_x = True
    default = "inf"


# Couplings
class KWZeeman(MuSpinCouplingKeyword):

    name = "zeeman"
    expr_size_bounds = (3, 3)
    block_size = 1
    _constants = {**_math_constants, **_phys_constants}

    def _default_args(self, i):
        args = {"i": int(i)}
        return args


class KWDipolar(MuSpinCouplingKeyword):

    name = "dipolar"
    block_size = 1
    expr_size_bounds = (3, 3)

    def _default_args(self, i, j):
        args = {"i": int(i), "j": int(j)}
        return args


class KWHyperfine(MuSpinCouplingKeyword):

    name = "hyperfine"
    block_size = 3
    expr_size_bounds = (3, 3)

    def _default_args(self, i, j=None):
        args = {"i": int(i), "j": int(j) if j is not None else None}
        return args


class KWQuadrupolar(MuSpinCouplingKeyword):

    name = "quadrupolar"
    block_size = 3
    expr_size_bounds = (3, 3)

    def _default_args(self, i):
        args = {
            "i": int(i),
        }
        return args


class KWDissipation(MuSpinCouplingKeyword):

    name = "dissipation"
    block_size = 1
    expr_size_bounds = (1, 1)

    def _default_args(self, i):
        args = {"i": int(i)}
        return args


# Fitting variables. This is a special case
class KWFittingVariables(MuSpinKeyword):

    name = "fitting_variables"
    block_size = 1
    expr_size_bounds = (1, 4)
    accept_range = True
    default = ""
    _constants = {**_math_constants, **_phys_constants}
    _validators = [
        lambda s: "Invalid value '{0}': variable name conflicts with a constant".format(
            s.name
        )
        if s.name in {**_math_constants, **_phys_constants}
        else ""
    ]

    def _store_values(self, block):

        variables = list(self._constants.keys())

        self._values = []

        for v in block:
            v = v[0].split(maxsplit=1)
            if len(v) == 0:
                return
            b = [v[0]]

            if len(v) == 2:
                b += [
                    LarkExpression(tk, variables=variables)
                    for tk in lark_tokenize(v[1])
                ]
                b[1:] = [expr.evaluate(**self._constants) for expr in b[1:]]

            b = FittingVariable(*b)

            self._values.append(b)


# Other fitting-related keywords
class KWFittingData(MuSpinExpandKeyword):

    name = "fitting_data"
    block_size = 1
    expr_size_bounds = (1, 2)
    accept_range = True
    default = ""
    _constants = {}
    _functions = {"load": np.loadtxt}


class KWFittingMethod(MuSpinKeyword):

    name = "fitting_method"
    block_size = 1
    accept_range = False
    default = "nelder-mead"
    _validators = [
        lambda s: "Invalid value {0}, accepted values {1}".format(
            s[0].lower(), ["nelder-mead", "lbfgs"]
        )
        if s[0].lower() not in ["nelder-mead", "lbfgs"]
        else ""
    ]


class KWFittingTolerance(MuSpinKeyword):

    name = "fitting_tolerance"
    block_size = 1
    accept_range = False
    default = "1e-3"
    _validators = [
        lambda s: "Invalid value '{0}', accepts only single float value".format(s[0])
        if not float(s[0])
        else "",
    ]


# Special configuration keyword
class KWExperiment(MuSpinKeyword):

    name = "experiment"
    block_size = 1
    accept_range = False
    default = ""


# Compile all KW classes into a single dictionary automatically
InputKeywords = {
    obj.name: obj
    for name, obj in inspect.getmembers(sys.modules[__name__])
    if (inspect.isclass(obj) and re.match("KW[0-9a-zA-Z]+", name))
}



"""spinop.py

Utility class to create and manipulate spin operators
"""

import numpy as np
from numbers import Number
from muspinsim.utils import Clonable
from scipy import sparse


def _mvals(I):
    return np.linspace(I, -I, int(2 * I + 1))


def _Sp(mvals):
    return 2 * np.diag(0.5 * (np.cumsum(2 * mvals)[:-1] ** 0.5), k=1) + 0.0j


def _Sm(mvals):
    return _Sp(mvals).T


def _Sx(mvals):
    o = _Sp(mvals)
    return 0.5 * (o + o.T)


def _Sy(mvals):
    o = _Sp(mvals)
    return 0.5j * (o.T - o)


def _Sz(mvals):
    return np.diag(mvals) + 0.0j


def _S0(mvals):
    return np.eye(len(mvals)) + 0.0j


class Hermitian(object):
    """A helper mixin for operators that are also Hermitian"""

    def __init__(self):

        if not self.is_hermitian:
            raise ValueError("Operator must be hermitian")

    def diag(self):

        """Diagonalise the operator

        Return eigenvalues and corresponding eigenvectors for this operator.

        Returns:
            (ndarray, ndarray) -- Eigenvalues and eigenvector matrix, as
                                  returned by numpy.linalg.eigh
        """

        try:
            dd = self._diagdata
            if np.all(dd["matrix"] == self._matrix):
                return dd["eigh"]
        except AttributeError:
            pass

        # need to convert this to use sparse matrices instead
        eigh = np.linalg.eigh(self._matrix.toarray())

        # setting k is important
        # too big will cause memory issues
        # too small = inaccurate results
        # eigval, eigvec = sparse.linalg.eigsh(self._matrix, k=self._matrix.shape[0]-2)
        # idx = eigval.argsort()
        # eigval = eigval[idx]
        # eigvec = eigvec[:,idx]
        # eigh = (eigval, eigvec)

        self._diagdata = {"matrix": self._matrix.copy(), "eigh": eigh}

        return eigh


class Operator(Clonable):
    def __init__(self, matrix, dim=None, hermtol=1e-6):
        """Create a Operator object

        Create an object representing a spin operator. These can
        be manipulated by e.g. multiplying them by a scalar or among themselves
        (equivalent to a dot product), or adding and subtracting them.

        Arguments:
            matrix {ndarray} -- Matrix describing the operator (must be a
                                square 2D array)

        Keyword Arguments:
            dim {(int,...)} -- Tuple of the dimensions of the operator. For example,
                               (2,2) corresponds to two 1/2 spins and a 4x4 matrix.
                               If not specified, it's taken from the size of
                               the matrix (default: {None})
            hermtol {float} -- Tolerance used to check for hermitianity of the
                               matrix (default: {1e-6})

        Raises:
            ValueError -- Any of the passed values are invalid
        """
        # use sparse matricies
        self._matrix = sparse.csr_matrix(matrix)

        if not (matrix.shape[0] == matrix.shape[1]):
            raise ValueError("Matrix passed to Operator must be square")

        if dim is None:
            dim = (matrix.shape[0],)
        elif np.prod(dim) != matrix.shape[0]:
            raise ValueError("Dimensions are not compatible with matrix")

        self._dim = tuple(dim)

        self._htol = hermtol

        super(Operator, self).__init__()

    @property
    def Is(self):
        return tuple((d - 1) / 2.0 for d in self._dim)

    @property
    def dimension(self):
        return self._dim

    @property
    def N(self):
        return np.prod(self._dim)

    @property
    def matrix(self):
        return self._matrix

    @property
    def is_hermitian(self):
        x = self._matrix.conjugate().T
        y = self._matrix
        return np.abs(y - x).max() < self._htol

    def dagger(self):
        """Return the transpose conjugate of this Operator

        Return the transpose conjugate of this Operator

        Returns:
            Operator -- Transpose conjugate of this operator
        """

        MyClass = self.__class__
        ans = MyClass.__new__(MyClass)
        ans._dim = tuple(self._dim)
        ans._matrix = self.matrix.conjugate().T

        return ans

    def __add__(self, x):

        if isinstance(x, Operator):

            if self.dimension != x.dimension:
                raise ArithmeticError(
                    "Can not add to Operators with different dimensions"
                )

            ans = self.clone()
            ans._matrix += x._matrix

            return ans

        elif isinstance(x, Number):

            ans = self.clone()
            ans._matrix += sparse.eye(ans._matrix.shape[0]) * x

            return ans

        raise TypeError("Unsupported operation for Operator")

    def __sub__(self, x):

        if isinstance(x, Operator):

            if self.dimension != x.dimension:
                raise ArithmeticError(
                    "Can not subtract to Operators with different dimensions"
                )

            ans = self.clone()
            ans._matrix -= x._matrix

            return ans

        elif isinstance(x, Number):

            return self + (-x)

        raise TypeError("Unsupported operation for Operator")

    def __mul__(self, x):

        if isinstance(x, Operator):

            if self.dimension != x.dimension:
                raise ArithmeticError(
                    "Can not multiply to Operators with different dimensions"
                )

            ans = self.clone()
            ans._matrix = ans._matrix.dot(x._matrix)

            return ans

        elif isinstance(x, Number):

            ans = self.clone()
            ans._matrix = ans._matrix.multiply(x)

            return ans

        raise TypeError("Unsupported operation for Operator")

    def __rmul__(self, x):

        if isinstance(x, Number):
            return self.__mul__(x)
        elif isinstance(x, Operator):
            return x * self

        raise TypeError("Unsupported operation for Operator")

    def __truediv__(self, x):

        if isinstance(x, Number):
            x = 1.0 / x
            return self * x

        raise TypeError("Unsupported operation for Operator")

    def __eq__(self, x):

        if not isinstance(x, Operator):
            return False

        if self.dimension != x.dimension:
            return False

        return np.all(self._matrix == x._matrix)

    def kron(self, x):
        """Tensor product between this and another Operator

        Performs a tensor product between this and another Operator,
        raising the overall rank of the tensor they represent.

        Arguments:
            x {Operator} -- Other operator

        Returns:
            Operator -- Result

        Raises:
            ValueError -- Thrown if x is not the right type of object
        """

        if not isinstance(x, Operator):
            raise ValueError("Can only perform Kronecker product with another Operator")

        # Doing it this way saves some time
        ans = self.__class__.__new__(self.__class__)
        ans._dim = self._dim + x._dim
        ans._matrix = sparse.kron(self._matrix, x._matrix, format="csr")

        return ans

    def hilbert_schmidt(self, x):
        """Hilbert-Schmidt product between this and another Operator


        Performs a Hilbert-Schmidt product between this and another Operator,
        that acts as an inner product.

        Arguments:
            x {Operator} -- Other operator

        Returns:
            number -- Result

        Raises:
            ValueError -- Thrown if x is not the right type of object
        """

        if not isinstance(x, Operator):
            raise ValueError(
                "Can only perform Hilbert-Schmidt product with another Operator"
            )

        if not x.dimension == self.dimension:
            raise ValueError(
                "Operators must have the same dimension to "
                "perform Hilbert-Schmidt product"
            )

        A = self.matrix
        B = x.matrix

        return A.conjugate().T.dot(B).trace().item()

    def basis_change(self, basis):
        """Return a version of this Operator with different basis

        Transform this Operator to use a different basis. The basis
        must be a matrix of orthogonal vectors. Passing as basis
        the eigenvectors of the operator will diagonalise it.

        Arguments:
            basis {ndarray} -- Basis to transform the operator to.

        Returns:
            Operator -- Basis transformed version of this operator
        """

        ans = self.clone()
        basis = sparse.csr_matrix(basis)
        x = basis.T.conjugate()
        ans._matrix = x.dot(ans._matrix).dot(basis)

        return ans


class SpinOperator(Operator):
    @classmethod
    def from_axes(self, Is=0.5, axes="x"):
        """Construct a SpinOperator from spins and axes

        Construct a SpinOperator from a list of spin values and directions. For
        example, Is=[0.5, 0.5] axes=['x', 'z'] will create a SxIz operator between
        two spin 1/2 particles.

        Keyword Arguments:
            Is {[number]} -- List of spins (must be half-integers). Can pass a
                             number if it's only one value (default: {0.5})
            axes {[str]} -- List of axes, can pass a single character if it's
                            only one value. Each value can be x, y, z, +, -,
                            or 0 (for the identity operator) (default: {'x'})

        Returns:
            SpinOperator -- Operator built according to specifications

        Raises:
            ValueError -- Any of the values passed is invalid
        """

        if isinstance(Is, Number):
            Is = [Is]

        if len(Is) != len(axes) or len(Is) == 0:
            raise ValueError("Arrays of moments and axes must have same length > 0")

        dim = tuple(int(2 * I + 1) for I in Is)
        matrices = []

        for I, axis in zip(Is, axes):

            if I % 0.5 or I < 0.5:
                raise ValueError("{0} is not a valid spin value".format(I))

            if not (axis in "xyz+-0"):
                raise ValueError("{0} is not a valid spin axis".format(axis))

            mvals = _mvals(I)

            o = {"x": _Sx, "y": _Sy, "z": _Sz, "+": _Sp, "-": _Sm, "0": _S0}[axis](
                mvals
            )

            matrices.append(o)

        M = matrices[0]
        for m in matrices[1:]:
            M = np.kron(M, m)

        return self(sparse.csr_matrix(M), dim=dim)


class DensityOperator(Operator):
    def __init__(self, matrix, dim=None):
        """Create a DensityOperator object

        Create an object representing a density operator. These can
        be manipulated by e.g. multiplying them by a scalar or among themselves
        (equivalent to a dot product), or adding and subtracting them.

        Arguments:
            matrix {ndarray} -- Matrix describing the operator (must be a
                                square hermitian 2D array and have non-zero
                                trace; will be normalised to have trace 1)

        Keyword Arguments:
            dim {(int,...)} -- Tuple of the dimensions of the operator. For example,
                               (2,2) corresponds to two 1/2 spins and a 4x4 matrix.
                               If not specified, it's taken from the size of
                               the matrix (default: {None})

        Raises:
            ValueError -- Any of the passed values are invalid
        """
        super(DensityOperator, self).__init__(matrix, dim)
        # Enforce unitarity

        tr = self._matrix.trace()

        if tr == 0:
            raise ValueError("Can not define a DensityOperator with zero trace")
        else:
            self.normalize()

        if not self.is_hermitian:
            raise ValueError("DensityOperator must be hermitian!")

    @classmethod
    def from_vectors(self, Is=0.5, vectors=[0, 0, 1], gammas=0):
        """Construct a density matrix state from real space vectors

        Construct a density matrix state by specifying a number of spins and
        real space directions. The state is initialised as the tensor product
        of independent spin states each pointing in the specified direction.
        A parameter gamma can be used to include decoherence effects and thus
        dampen or zero out all off-diagonal elements.

        Keyword Arguments:
            Is {[number]} -- List of spins (must be half-integers). Can pass a
                             number if it's only one value (default: {0.5})
            vectors {[ndarray]} -- List of vectors. Can pass a single 3D vector
                                   if it's only one value (default: {[0, 0, 1]})
            gammas {[number]} -- List of gamma factors. Can pass a single number
                                 if it's only one value. All off-diagonal
                                 elements for each corresponding density matrix
                                 will be multiplied by 1-gamma. (default: {0})

        Returns:
            DensityOperator -- The composite density operator

        Raises:
            ValueError -- Any of the passed values are invalid
        """

        if isinstance(Is, Number):
            Is = [Is]

        if len(np.array(vectors).shape) == 1:
            vectors = [vectors]

        if isinstance(gammas, Number):
            gammas = [gammas]

        if len(Is) != len(vectors) or len(Is) != len(gammas) or len(Is) == 0:
            raise ValueError(
                "Arrays of moments, axes and gammas must have same length > 0"
            )

        dim = tuple(int(2 * I + 1) for I in Is)
        matrices = []

        for I, vec, gamma in zip(Is, vectors, gammas):

            if I % 0.5 or I < 0.5:
                raise ValueError("{0} is not a valid spin value".format(I))

            if not len(vec) == 3:
                raise ValueError("{0} is not a valid 3D vector".format(vec))

            if gamma < 0 or gamma > 1:
                raise ValueError("{0} is not a valid gamma value".format(gamma))

            mvals = _mvals(I)

            S = [_Sx(mvals), _Sy(mvals), _Sz(mvals)]

            o = sum([S[i] * vec[i] for i in range(3)])

            evals, evecs = np.linalg.eigh(o)

            psi = evecs[:, np.argmax(evals)]

            m = psi[:, None] * psi[None, :].conj()
            m *= (1 - gamma) * np.ones(m.shape) + gamma * np.eye(m.shape[0])

            matrices.append(m)

        M = matrices[0]
        for m in matrices[1:]:
            M = np.kron(M, m)

        return self(sparse.csr_matrix(M), dim=dim)

    @property
    def trace(self):
        return self._matrix.trace()

    def normalize(self):
        """Normalize this DensityOperator to have trace equal to one."""
        self._matrix = self._matrix.multiply(1 / self.trace)

    def partial_trace(self, tracedim=[]):
        """Perform a partial trace operation

        Perform a partial trace over the specified dimensions and return the
        resulting DensityOperator.

        Keyword Arguments:
            tracedim {[int]} -- Indices of dimensions to perform the partial
                                trace over (default: {[]})

        Returns:
            DensityOperator -- Operator with partial trace
        """

        dim = list(self._dim)
        tdim = list(sorted(tracedim))

        m = self._matrix.toarray().reshape(dim + dim)

        while len(tdim) > 0:
            td = tdim.pop(-1)
            # Trace along tdim
            m = np.trace(m, axis1=td, axis2=td + len(dim))
            dim.pop(td)  # Reduce dimension accordingly

        return DensityOperator(m, dim)

    def expectation(self, operator):
        """Compute expectation value of one operator

        Compute expectation value of an operator over the state defined by
        this DensityOperator.

        Arguments:
            operator {SpinOperator} -- Operator to compute the expectation
                                       value of

        Returns:
            number -- Expectation value

        Raises:
            TypeError -- The argument isn't a SpinOperator
            ValueError -- The operator isn't compatible with this one
        """

        if not isinstance(operator, SpinOperator):
            raise TypeError("Argument must be a SpinOperator")

        if not operator.dimension == self.dimension:
            raise ValueError(
                "SpinOperator and DensityOperator do not have compatible dimensions"
            )

        return operator.matrix.multiply(self.matrix.T).sum()


class SuperOperator(Operator):
    def __init__(self, matrix, dim=None):

        n = matrix.shape[0] ** 0.5

        if int(n) != n:
            raise ValueError("Matrix rank should be a perfect square")

        if dim is None:
            dim = (n, n)

        super(SuperOperator, self).__init__(matrix, dim)

    @classmethod
    def left_multiplier(self, operator):
        """Create a SuperOperator that performs a left multiplication

        Create a superoperator L from an operator O such that

        L*rho = O*rho

        Arguments:
            operator {Operator} -- Operator O

        Returns:
            SuperOperator -- SuperOperator L
        """

        m = operator.matrix
        d = operator.dimension

        M = sparse.kron(m, sparse.eye(m.shape[0]))

        return self(M, d + d)

    @classmethod
    def right_multiplier(self, operator):
        """Create a SuperOperator that performs a right multiplication

        Create a superoperator L from an operator O such that

        L*rho = rho*O

        Arguments:
            operator {Operator} -- Operator O

        Returns:
            SuperOperator -- SuperOperator L
        """

        m = operator.matrix
        d = operator.dimension

        M = sparse.kron(sparse.eye(m.shape[0]), m.T)

        return self(M, d + d)

    @classmethod
    def commutator(self, operator):
        """Create a SuperOperator that performs a commutation

        Create a superoperator L from an operator O such that

        L*rho = O*rho-rho*O

        Arguments:
            operator {Operator} -- Operator O

        Returns:
            SuperOperator -- SuperOperator L
        """

        return self.left_multiplier(operator) - self.right_multiplier(operator)

    @classmethod
    def anticommutator(self, operator):
        """Create a SuperOperator that performs an anticommutation

        Create a superoperator L from an operator O such that

        L*rho = O*rho+rho*O

        Arguments:
            operator {Operator} -- Operator O

        Returns:
            SuperOperator -- SuperOperator L
        """

        return self.left_multiplier(operator) + self.right_multiplier(operator)

    @classmethod
    def bracket(self, operator):
        """Create a SuperOperator that performs a basis change

        Create a superoperator L from an operator O such that

        L*rho = O*rho*O^

        where O^ is the conjugate transpose of O.

        Arguments:
            operator {Operator} -- Operator O

        Returns:
            SuperOperator -- SuperOperator L
        """

        m = operator.matrix
        d = operator.dimension

        M = sparse.kron(m, m.conjugate())

        return self(M, d + d)

    def __add__(self, x):

        if isinstance(x, SuperOperator) or isinstance(x, Number):
            return super(SuperOperator, self).__add__(x)

        raise TypeError("Unsupported operation for SuperOperator")

    def __sub__(self, x):

        if isinstance(x, SuperOperator) or isinstance(x, Number):
            return super(SuperOperator, self).__sub__(x)

        raise TypeError("Unsupported operation for SuperOperator")

    def __mul__(self, x):

        if isinstance(x, SuperOperator) or isinstance(x, Number):
            return super(SuperOperator, self).__mul__(x)
        elif isinstance(x, Operator):
            # Vectorize x
            m = x.matrix
            s = m.shape
            m = m.toarray().reshape((-1,))
            m = self.matrix.dot(m).reshape(s)
            return Operator(m, x.dimension)

        raise TypeError("Unsupported operation for Operator")

    def __rmul__(self, x):

        if isinstance(x, Number):
            return self.__mul__(x)
        elif isinstance(x, SuperOperator):
            return x * self

        raise TypeError("Unsupported operation for Operator")

    def __truediv__(self, x):

        if isinstance(x, Number):
            x = 1.0 / x
            return self * x

        raise TypeError("Unsupported operation for Operator")

    def __eq__(self, x):

        if not isinstance(x, Operator):
            return False

        if self.dimension != x.dimension:
            return False

        return np.all(self._matrix == x._matrix)



"""fitting.py

A class that takes care of runs where the goal is to fit some given data"""

import os
import numpy as np
from scipy.optimize import minimize

from muspinsim.input import MuSpinInput
from muspinsim.simconfig import MuSpinConfig
from muspinsim.experiment import ExperimentRunner
from muspinsim.mpi import mpi_controller as mpi


class FittingRunner(object):
    def __init__(self, inpfile: MuSpinInput):
        """Initialise a FittingRunner object

        Initialise an object to run a parallelised fitting calculation.
        The object only needs a MuSpinInput object defining the input file,
        and only on the root node it's important that this object has the file
        contents loaded. Everything else will be synchronised by broadcasting.

        Arguments:
            inpfile {MuSpinInput} -- Input file contents

        """

        self._input = inpfile
        self._runner = None

        # Identify variables
        self._fitinfo = inpfile.fitting_info

        if mpi.is_root:
            self._ytarg = self._fitinfo["data"][:, 1]

            variables = inpfile.variables

            if len(variables) == 0:
                # Should never happen, but...
                raise ValueError(
                    "MuSpinInput passed to FittingRunner has no variables to fit"
                )

            self._xnames = tuple(sorted(variables.keys()))  # Order is important!
            self._x = np.array([variables[n].value for n in self._xnames])
            self._xbounds = [variables[n].bounds for n in self._xnames]

            mpi.broadcast_object(self, ["_ytarg", "_x", "_xbounds", "_xnames"])
        else:
            mpi.broadcast_object(self, ["_ytarg", "_x", "_xbounds", "_xnames"])

        self._done = False
        self._sol = None

    def run(self, name=None, path="."):
        """Run a fitting calculation using Scipy, and returns the solution

        Returns:
            sol {scipy.OptimizeResult} -- The result of the optimisation.
        """

        if mpi.is_root:

            # Get the correct string for the method
            method = {"nelder-mead": "nelder-mead", "lbfgs": "L-BFGS-B"}[
                self._fitinfo["method"].lower()
            ]

            self._sol = minimize(
                self._targfun,
                self._x,
                method=method,
                tol=self._fitinfo["rtol"],
                bounds=self._xbounds,
            )

            self._done = True
            mpi.broadcast_object(self, ["_x", "_done"])

            mpi.broadcast_object(self, ["_sol"])

            # And now save the last result
            self._runner.config.save_output(name=name, path=path)
        else:
            while not self._done:
                self._targfun(self._x)
            # Receive solution
            mpi.broadcast_object(self, ["_sol"])

        return self._sol

    def _targfun(self, x):

        self._x = x
        # Synchronize across nodes
        mpi.broadcast_object(self, ["_x", "_done"])

        if self._done:
            # Child nodes will learn it here
            return

        vardict = dict(zip(self._xnames, self._x))
        self._runner = ExperimentRunner(self._input, variables=vardict)
        y = self._runner.run()

        if mpi.is_root:
            # Compare with target data
            err = np.average(np.abs(y - self._ytarg))
            return err

    def write_report(self, fname=None, path="./"):
        """Write a report file with the contents of the fitting optimisation.

        Write a human readable report summing up the results of the
        optimisation, including values found for the variables and tolerance
        achieved.

        Keyword Arguments:
            fname {str} -- Name to give to the report. If None, use the input
                           file's given name as base (default: None)
            path {str} -- Path at which to write the report (default: ./)
        """

        if mpi.is_root:

            # Final variable values
            variables = dict(zip(self._xnames, self._sol["x"]))
            config = MuSpinConfig(self._input.evaluate(**variables))

            if fname is None:
                fname = config.name + "_fit_report.txt"

            with open(os.path.join(path, fname), "w") as f:
                f.write("Fitting process for {0} completed\n".format(config.name))
                f.write("Success achieved: {0}\n".format(self._sol["success"]))
                if not self._sol["success"]:
                    f.write("   Message: {0}\n".format(self._sol["message"]))

                f.write(
                    "Final absolute error <|f-f_targ|>: "
                    "{0}\n".format(self._sol["fun"])
                )
                f.write("Number of simulations: {0}\n".format(self._sol["nfev"]))
                f.write("Number of iterations: {0}\n".format(self._sol["nit"]))

                f.write("\n" + "=" * 20 + "\n")
                f.write("\nValues found for fitting variables:\n\n")
                for name, val in variables.items():
                    f.write("\t{0} = {1}\n".format(name, val))

    @property
    def solution(self):
        return self._sol



"""utils.py

Utility functions and classes"""

from copy import deepcopy
from collections.abc import Iterable

import numpy as np
from ase.quaternions import Quaternion
from soprano.calculate.powder import ZCW


class Clonable(object):
    """A helper class; any object inheriting
    from this will have a .clone method that copies it easily."""

    def clone(self):

        MyClass = self.__class__
        copy = MyClass.__new__(MyClass)
        copy.__dict__ = deepcopy(self.__dict__)

        return copy


def deepmap(func, obj):
    """Deep traverse obj, and apply func to each of its non-Iterable
    elements"""

    if isinstance(obj, Iterable):
        return [deepmap(func, x) for x in obj]
    else:
        return func(obj)


def zcw_gen(N, mode="sphere"):
    N = int(N)
    pwd = ZCW(mode)
    return pwd.get_orient_angles(N)[0]


def eulrange_gen(N):
    # Generate a range of Euler angles and related weights
    N = int(N)  # Just making sure
    a = np.linspace(0, 2 * np.pi, N)
    b = np.linspace(0, np.pi, N + 2)[1:-1]
    c = np.linspace(0, 2 * np.pi, N)

    a, b, c = np.array(np.meshgrid(a, b, c)).reshape((3, -1))
    w = np.sin(b)

    return np.array([a, b, c, w]).T


def quat_from_polar(theta, phi):
    """Make a Quaternion from two polar angles

    Make a Quaternion from only two polar angles. This only makes the new Z'
    axis have polar angles theta and phi w.r.t. the old system, but does not
    care or guarantee anything about the other two angles.

    Arguments:
        theta {float} -- Zenithal angle
        phi {float} -- Azimuthal angle

    Returns:
        q {ase.Quaternion} -- Quaternion
    """

    return Quaternion.from_euler_angles(phi, theta, phi, "zyz")


def get_xy(z):
    """Make two axes x and y orthogonal and correctly oriented with respect
    to an axis z"""

    zn = np.linalg.norm(z)
    if zn == 0:
        raise ValueError("Can not find X and Y for null Z vector")
    z = np.array(z) / zn

    if z[0] == 0 and z[1] == 0:
        x = np.array([1.0, 0, 0])
        y = np.array([0, z[2], 0])
    else:
        x = np.array([z[1], -z[0], 0.0]) / (z[0] ** 2 + z[1] ** 2) ** 0.5
        y = np.array(
            [
                z[1] * x[2] - z[2] * x[1],
                z[2] * x[0] - z[0] * x[2],
                z[0] * x[1] - z[1] * x[0],
            ]
        )

    return x, y



import numpy as np

from numbers import Number
from muspinsim.spinop import DensityOperator, SpinOperator


def validate_times(times):
    """Validates the 'times' parameter for 'evolve' and 'integrate_decaying'
       methods in the  Hamiltonian, CelioHamiltonian and Linbladian classes

    Arguments:
       times {ndarray} -- Times to compute the evolution for, in microseconds

    Raises:
       ValueError -- Invalid values of times
    """
    if not isinstance(times, np.ndarray):
        raise TypeError("times must be an array of values in microseconds")
    if len(times.shape) != 1:
        raise ValueError("times must be an array of values in microseconds")


def validate_evolve_params(rho0, times, operators):
    """Validates the required parameters for 'evolve' methods
       in the Hamiltonian, CelioHamiltonian and Linbladian classes

    Arguments:
        rho0 {DensityOperator} -- Initial state
        times {ndarray} -- Times to compute the evolution for, in microseconds
        operators {[SpinOperator]} -- List of SpinOperators to compute the
                                      expectation values of at each step.
    Raises:
        TypeError -- Invalid operators
        ValueError -- Invalid values of times or operators
    """
    if not isinstance(rho0, DensityOperator):
        raise TypeError("rho0 must be a valid DensityOperator")

    validate_times(times)

    if not all([isinstance(o, SpinOperator) for o in operators]):
        raise ValueError(
            "operators must be a SpinOperator or a list of SpinOperator objects"
        )


def validate_integrate_decaying_params(rho0, tau, operators):
    """Validates the required parameters for 'integrate_decaying' methods
       in the Hamiltonian and Linbladian classes

    Arguments:
        rho0 {DensityOperator} -- Initial state
        tau {float} -- Decay time, in microseconds
        operators {list} -- Operators to compute the expectation values
                            of

    Raises:
        TypeError -- Invalid operators
        ValueError -- Invalid values of tau or operators
    """
    if not isinstance(rho0, DensityOperator):
        raise TypeError("rho0 must be a valid DensityOperator")

    if not (isinstance(tau, Number) and np.isreal(tau) and tau > 0):
        raise ValueError("tau must be a real number > 0")

    if not all([isinstance(o, SpinOperator) for o in operators]):
        raise ValueError(
            "operators must be a SpinOperator or a list of SpinOperator objects"
        )



"""celio.py

A class for handling the computation of hamiltonian for Celio's method via a
trotter expansion

See Phys. Rev. Lett. 56 2720 (1986)
"""

from dataclasses import dataclass
import itertools
import logging
from typing import List
import numpy as np
from scipy import sparse
from qutip import Qobj

from muspinsim.cpp import (
    Celio_EvolveContrib,
    celio_evolve,
)
from muspinsim.spinop import SpinOperator
from muspinsim.validation import validate_evolve_params, validate_times


@dataclass
class CelioHContrib:
    """
    Stores a Hamiltonian contribution term for use with Celio's method

    Arguments:
            matrix {matrix} -- Sparse matrix representing a contribution to the
                               hamiltonian
            other_dimension {int} -- Defines the product of the matrix sizes of any
                                     remaining spins that are not included in this
                                     hamiltonian contribution
            spin_order {[int]} -- Defines the order in which spins are included for
                                  this term. Its needed when constructing the
                                  contribution to the trotter hamiltonian after the
                                  matrix exponential
            spin_dimensions {[int]} -- Defines the size of the matrices involved
                                       in the kronecker products that make up this
                                       contribution to the Hamiltonian
    """

    matrix: sparse.csr_matrix
    other_dimension: int
    spin_order: List[int]
    spin_dimensions: List[int]


class CelioHamiltonian:
    def __init__(self, terms, k, spinsys):
        """Create a CelioHamiltonian

        Create a CelioHamiltonian for applying Celio's method

        Arguments:
            terms {[InteractionTerm]} -- Interaction terms that will form part of the
                                         Trotter expansion
            k {int} -- Factor to be used in the Trotter expansion
            spinsys {SpinSystem} -- SpinSystem required for computing the time evolution
        """
        self._terms = terms
        self._k = k
        self._spinsys = spinsys

    def __add__(self, x):
        return CelioHamiltonian(self._terms + x._terms, self._k, self._spinsys)

    def _calc_H_contribs(self) -> List[CelioHContrib]:
        """Calculates and returns the Hamiltonian contributions required for Celio's
           method

        Returns the hamiltonian contributions defined by this system of spins and the
        given interactions. In general these are split up per group of indices defined
        in interactions to minimise the need of matrix exponentials.

        Returns:
            H_contribs {[CelioHContrib]} -- List of matrices representing contributions
                                            to the total system hamiltonians referred to
                                            in Celio's method as H_i
        """

        spin_indices = range(0, len(self._spinsys.spins))

        H_contribs = []

        for i in spin_indices:
            # Only want to include each interaction once, will make the choice here to
            # only add it to the H_i for the first particle listed in the interactions

            # Find the terms that have the current spin as its first or only index
            spin_ints = [term for term in (self._terms) if i == term.indices[0]]

            # List of spin indices not included here
            other_spins = list(range(0, len(self._spinsys.spins)))
            other_spins.remove(i)

            # Only include necessary terms
            if len(spin_ints) != 0:
                # Sum matrices with the same indices so we avoid lots of matrix
                # exponentials
                for indices, group in itertools.groupby(
                    spin_ints, lambda term: term.indices
                ):

                    grouped_spin_ints = list(group)
                    H_contrib = np.sum([term.matrix for term in grouped_spin_ints])

                    # Find indices of spins not involved in the current interactions
                    uninvolved_spins = other_spins.copy()
                    for term in grouped_spin_ints:
                        for j in term.indices:
                            if j in uninvolved_spins:
                                uninvolved_spins.remove(j)

                    other_dimension = np.product(
                        [self._spinsys.dimension[j] for j in uninvolved_spins]
                    )

                    # Detect Quadrupolar terms which will have the same index twice
                    indices = list(indices)
                    if len(indices) == 2 and indices[0] == indices[1]:
                        # Only include one of them for the ordering
                        indices.pop()

                    # Order in which kronecker products will be performed in Celio's
                    # method and the corresponding dimensions
                    spin_order = indices + uninvolved_spins
                    spin_dimensions = [self._spinsys.dimension[i] for i in spin_order]

                    H_contribs.append(
                        CelioHContrib(
                            H_contrib,
                            other_dimension,
                            spin_order,
                            spin_dimensions,
                        )
                    )

        return H_contribs

    def _calc_trotter_evol_op_contribs(self, time_step, cpp):
        """Calculates and returns the contributions to the Trotter expansion of
        the time evolution operator computed from the Hamiltonian contributions

        Arguments:
            time_step {float} -- Timestep that will be used during the evolution
            cpp {boolean} -- When true will calculate ready for use with the C++
                             version

        Returns:
            evol_op_contribs {[matrix]} -- Contributions to the Trotter expansion
                                           of the evolution operator
        """

        H_contribs = self._calc_H_contribs()

        evol_op_contribs = []

        for H_contrib in H_contribs:
            # The matrix is currently stored in csr format, but expm wants it
            # in csc so convert here
            evol_op_contrib = sparse.linalg.expm(
                -2j * np.pi * H_contrib.matrix.tocsc() * time_step / self._k
            ).tocsr()

            if cpp:
                # C++ version - No kronecker products required - just an array
                # of indices
                evol_op_contribs.append(
                    Celio_EvolveContrib(
                        evol_op_contrib.toarray(),
                        int(H_contrib.other_dimension),
                        np.transpose(
                            np.arange(
                                np.product(self._spinsys.dimension), dtype=np.uint64
                            ).reshape(self._spinsys.dimension),
                            axes=H_contrib.spin_order,
                        ).flatten(),
                    )
                )
            else:
                # Python version - requires kronecker products
                if H_contrib.other_dimension > 1:
                    evol_op_contrib = sparse.kron(
                        evol_op_contrib,
                        sparse.identity(H_contrib.other_dimension, format="csr"),
                    )

                # For particle interactions that are not neighbors we must use
                # a swap gate
                qtip_obj = Qobj(
                    inpt=evol_op_contrib,
                    dims=[H_contrib.spin_dimensions, H_contrib.spin_dimensions],
                )
                # Order we need to permute in order to obtain the same order
                # as was given in the input
                permute_order = np.argsort(H_contrib.spin_order)

                qtip_obj = qtip_obj.permute(permute_order)
                evol_op_contrib = qtip_obj.data

                evol_op_contribs.append(evol_op_contrib)

        return evol_op_contribs

    def evolve(self, rho0, times, operators=[]):
        """Time evolution of a state under this Hamiltonian

        Perform an evolution of a state described by a DensityOperator under
        this Hamiltonian and return either a sequence of DensityOperators or
        a sequence of expectation values for given SpinOperators.

        Arguments:
            rho0 {DensityOperator} -- Initial state
            times {ndarray} -- Times to compute the evolution for, in microseconds

        Keyword Arguments:
            operators {[SpinOperator]} -- List of SpinOperators to compute the
                                          expectation values of at each step.
                                          If omitted, the states' density
                                          matrices will be returned instead
                                           (default: {[]})

        Returns:
            [DensityOperator | ndarray] -- DensityOperators or expectation values

        Raises:
            TypeError -- Invalid operators
            ValueError -- Invalid values of times or operators
            RuntimeError -- Hamiltonian is not hermitian
        """

        times = np.array(times)
        if isinstance(operators, SpinOperator):
            operators = [operators]

        validate_evolve_params(rho0, times, operators)

        if len(self._terms) == 0:
            raise ValueError("No interaction terms to evolve")

        time_step = times[1] - times[0]
        rho0 = rho0.matrix

        # Time evolution step that will modify the trotter_hamiltonian below
        evol_op = (
            np.product(self._calc_trotter_evol_op_contribs(time_step, False)) ** self._k
        )

        total_evol_op = sparse.identity(evol_op.shape[0], format="csr")

        mat_density = evol_op.getnnz() / np.prod(evol_op.shape)

        if mat_density >= 0.08:
            logging.warning(
                "Matrix density is %s >= 0.08. Using Celio's method "
                "without random initial states will be slow.",
                mat_density,
            )
            # Matrix products with trotter_hamiltonian_dt is very likely to be slower
            # with sparse matrices than dense

            # We can still save some memory over Hamiltonian's evolve method at the
            # cost of performance by using dense matrices for trotter_hamiltonian
            # trotter_hamiltonian_dt but the improvement is minimal and as the problem
            # gets bigger the reduction in memory usage decreases and increase in time
            # increases so does not appear worth it

        # Avoid using append as assignment should be faster
        results = np.zeros((times.shape[0], len(operators)), dtype=np.complex128)

        # Obtain transpose of operators
        operatorsT = np.array([o.matrix.T for o in operators])

        if len(operators) > 0:
            # Compute expectation values one at a time
            for i in range(times.shape[0]):
                # Evolve rho0
                rho = total_evol_op.conj().T * (rho0 * total_evol_op)

                # When passing multiple operators we want to return results for each
                for j, operatorT in enumerate(operatorsT):
                    # This element wise multiplication then sum gives the equivalent
                    # as the trace of the matrix product (without the transpose) and
                    # and is faster
                    results[i][j] = np.sum(
                        np.sum(rho.multiply(operatorT), axis=1), axis=0
                    )

                # Evolution step
                total_evol_op = total_evol_op * evol_op

        return results

    def _compute_psi(self, mu_psi, half_dim):
        """Computes a random initial muon state

        The muon index should be first for this to work correctly, otherwise
        the order of the kronecker products will need be be modified.

        Arguments:
            mu_psi {ndarray} -- Eigenvector from sigma_mu + eye(2)
            half_dim {int} -- Half the total dimension of the system

        Returns:
            [ndarray] -- Randomised initial state for the system with a shape
                         (mu_psi.shape[0] * half_dim, 1)
        """
        psi0 = np.exp(2j * np.pi * np.random.rand(half_dim))
        psi = np.kron(mu_psi.T, psi0)

        # Normalise
        psi = psi * (1.0 / np.sqrt(half_dim))

        # Likely dense, faster to use numpy array
        return psi.T

    def fast_evolve(self, sigma_mu, times, averages, cpp=True):
        """Time evolution of spin states under this Hamiltonian

        Perform the time evolution of a randomised initial spin state under
        this Hamiltonian and return a sequence of expectation values for the
        muon polarisation at the requested times

        Arguments:
            sigma_mu {ndarray} -- Linear combination of Pauli spin matrices in
                                  the direction of the muon
            times {ndarray} -- Times to compute the evolution for, in
                               microseconds
            averages {int} -- Number of averages to compute
            cpp {boolean} -- When true will calculate ready for use with the
                             C++ version

        Returns:
            [ndarray] -- Expectation values

        Raises:
            TypeError -- Invalid operators
            ValueError -- Invalid values of times
            ValueError -- Invalid value of averages
            ValueError -- If there are no interaction terms for the evolution
            ValueError -- If the muon is not first in the system
            RuntimeError -- Hamiltonian is not hermitian
        """

        times = np.array(times)

        validate_times(times)

        if averages <= 0:
            raise ValueError("averages must be a positive integer")

        if len(self._terms) == 0:
            raise ValueError("No interaction terms to evolve")

        # Due to computation of of psi we assume the muon is first in
        # the system so ensure this is the case here, otherwise
        # we need to change the order of kronecker products when computing
        # it and this would be slower anyway
        if self._spinsys.muon_index != 0:
            raise ValueError(
                "Muon must be the first spin in the system in order to use "
                "the fast Celio method"
            )

        time_step = times[1] - times[0]

        # Obtain spin up and down states, and select the one with
        # the eigenvalue +1 while trying to avoid precision issues
        evals, evecs = np.linalg.eig(sigma_mu + np.eye(2))
        mu_psi = evecs[:, 1] if evals[1] > 0.1 else evecs[:, 0]

        dimension = np.product(self._spinsys.dimension)
        half_dim = int(dimension / 2)

        if cpp:
            return self._fast_evolve_cpp(
                sigma_mu, times, averages, mu_psi, half_dim, time_step
            )
        else:
            return self._fast_evolve_python(
                sigma_mu, times, averages, mu_psi, half_dim, time_step
            )

    def _fast_evolve_python(
        self, sigma_mu, times, averages, mu_psi, half_dim, time_step
    ):
        """Time evolution of spin states under this Hamiltonian

        Perform the time evolution of a randomised initial spin state under
        this Hamiltonian and return a sequence of expectation values for the
        muon polarisation at the requested times

        Arguments:
            sigma_mu {matrix} -- Spin matrix of the muon
            times {ndarray} -- Times to compute the evolution for, in microseconds
            averages {int} -- Number of averages to compute
            mu_psi {ndarray} -- Eigenvector from sigma_mu + eye(2)
            half_dim {int} -- Half the total dimension of the system
            time_step {float} -- Difference between subsequent times

        Returns:
            [ndarray] -- Expectation values
        """

        # Time evolution step that will modify the trotter_hamiltonian below
        evol_op_contribs = self._calc_trotter_evol_op_contribs(time_step, False)

        operator = sparse.kron(sigma_mu, sparse.identity(half_dim, format="csr")).T

        # Avoid using append as assignment should be faster
        results = np.zeros(times.shape[0], dtype=np.complex128)

        avg_factor = 1.0 / averages

        for _ in range(averages):
            psi = self._compute_psi(mu_psi, half_dim)

            # Compute expectation values one at a time
            for i in range(times.shape[0]):
                # Use @ symbol to avoid confusion when using numpy arrays
                results[i] += psi.conj().T @ (operator @ psi)

                # Evolution step
                for _ in range(self._k):
                    for evol_op_contrib in evol_op_contribs:
                        psi = evol_op_contrib * psi

        # Divide by 2 as by convention rest of muspinsim gives results between
        # 0.5 and -0.5
        return results * avg_factor * 0.5

    def _fast_evolve_cpp(self, sigma_mu, times, averages, mu_psi, half_dim, time_step):
        """Parallelised time evolution of a state under this Hamiltonian

        Perform the time evolution of a randomised initial spin state under
        this Hamiltonian and return a sequence of expectation values for the
        muon polarisation at the requested times

        Arguments:
            sigma_mu {matrix} -- Spin matrix of the muon
            times {ndarray} -- Times to compute the evolution for, in microseconds
            averages {int} -- Number of averages to compute
            mu_psi {ndarray} -- Eigenvector from sigma_mu + eye(2)
            half_dim {int} -- Half the total dimension of the system
            time_step {float} -- Difference between subsequent times

        Returns:
            [ndarray] -- Expectation values
        """

        # Time evolution step that will modify the trotter_hamiltonian below
        evol_contribs = self._calc_trotter_evol_op_contribs(time_step, True)

        # Avoid using append as assignment should be faster
        results = np.zeros(times.shape[0], dtype=np.float64)

        avg_factor = 1.0 / averages

        sigma_mu = sigma_mu.toarray()

        for _ in range(averages):
            psi = self._compute_psi(mu_psi, half_dim)

            # Compute expectation values
            celio_evolve(
                times.shape[0],
                psi,
                sigma_mu,
                half_dim,
                self._k,
                evol_contribs,
                results,
            )

        # Divide by 2 as by convention rest of muspinsim gives results between
        # 0.5 and -0.5
        return results * avg_factor * 0.5

    def integrate_decaying(self, rho0, tau, operators=[]):
        """Called to integrate one or more expectation values in time with decay

        Raises:
            NotImplementedError
        """
        raise NotImplementedError(
            "integrate_decaying is not implemented for Celio's method"
        )



"""spinsys.py

A class to hold a given spin system, defined by specific nuclei
"""

import logging

import numpy as np
from numbers import Number
import scipy.constants as cnst
from scipy import sparse
from qutip import sigmax, sigmay, sigmaz

from muspinsim.celio import CelioHamiltonian
from muspinsim.utils import Clonable
from muspinsim.spinop import SpinOperator
from muspinsim.hamiltonian import Hamiltonian
from muspinsim.lindbladian import Lindbladian
from muspinsim.constants import gyromagnetic_ratio, spin, quadrupole_moment, EFG_2_MHZ


class InteractionTerm(Clonable):
    def __init__(self, spinsys, indices=[], tensor=0, label=None):

        self._spinsys = spinsys
        self._indices = np.array(indices)
        self._tensor = np.array(tensor)
        self._label = "Term" if label is None else label

        if np.any(np.array(self._tensor.shape) != 3):
            raise ValueError("Tensor is not fully three-dimensional")

        self._recalc_operator()

    def _recalc_operator(self):

        total_op = None
        d = len(self._tensor.shape)

        if d > 0:
            index_tuples = np.indices(self._tensor.shape).reshape((d, -1)).T
        else:
            index_tuples = [[]]

        for ii in index_tuples:
            # Detect double terms with same index - e.g. quadrupole terms
            if len(self._indices) == 2 and self._indices[0] == self._indices[1]:
                # Here use matrix multiplication for the terms
                op = (
                    self._spinsys.operator(
                        {self.indices[0]: ["xyz"[ii[0]], "xyz"[ii[1]]]},
                        include_only_given=self._spinsys.celio_k,
                    )
                    * self._tensor[tuple(ii)]
                )
            else:
                # Here use kronecker products when there are more than one term
                op = (
                    self._spinsys.operator(
                        {ind: "xyz"[ii[i]] for i, ind in enumerate(self._indices)},
                        include_only_given=self._spinsys.celio_k,
                    )
                    * self._tensor[tuple(ii)]
                )

            if total_op is None:
                total_op = op
            else:
                total_op += op

        self._operator = total_op

    @property
    def label(self):
        return self._label

    @property
    def indices(self):
        return tuple(self._indices)

    @property
    def tensor(self):
        return np.array(self._tensor)

    @property
    def operator(self):
        return self._operator.clone()

    @property
    def matrix(self):
        return self._operator.matrix

    def __repr__(self):
        return self.label


class SingleTerm(InteractionTerm):
    def __init__(self, spinsys, i, vector, label="Single"):

        super(SingleTerm, self).__init__(spinsys, [i], vector, label)

    @property
    def i(self):
        return self._indices[0]

    def rotate(self, rotmat):

        R = np.array(rotmat)
        v = self._tensor
        v = np.dot(v, R.T)

        rt = SingleTerm(self._spinsys, self.i, v, self._label)

        return rt

    def __repr__(self):
        return "{0} {{ S_{1} * {2} }}".format(self._label, self.i, self._tensor)


class DoubleTerm(InteractionTerm):
    def __init__(self, spinsys, i, j, matrix, label="Double"):

        super(DoubleTerm, self).__init__(spinsys, [i, j], matrix, label)

    @property
    def i(self):
        return self._indices[0]

    @property
    def j(self):
        return self._indices[1]

    def rotate(self, rotmat):

        R = np.array(rotmat)
        M = self._tensor
        M = np.linalg.multi_dot([R, M, R.T])

        rt = DoubleTerm(self._spinsys, self.i, self.j, M, self._label)

        return rt

    def __repr__(self):
        return "{0} {{ S_{1} * [{2} {3} {4}] * S_{5} }}".format(
            self._label, self.i, *self._tensor, self.j
        )


class DissipationTerm(Clonable):
    def __init__(self, operator, gamma=0.0):

        self._op = operator
        self._g = gamma

    @property
    def operator(self):
        return self._op

    @property
    def gamma(self):
        return self._g

    @property
    def tuple(self):
        return (self._op, self._g)


class SpinSystem(Clonable):
    def __init__(self, spins=[], celio_k=0):
        """Create a SpinSystem object

        Create an object representing a system of particles with spins (muons,
        electrons and atomic nuclei) and holding their operators.

        Keyword Arguments:
            spins {list} -- List of symbols representing the various particles.
                            Each element can be 'e' (electron), 'mu' (muon) a
                            chemical symbol, or a (str, int) tuple with a
                            chemical symbol and an isotope (default: {[]})
            celio_k {int} -- Factor for the Trotter approximation if Celio's
                             method is to be used. When this is 0, Celio's
                             method is not used.
        """

        gammas = []
        Qs = []
        Is = []
        operators = []

        for s in spins:
            if isinstance(s, tuple):
                el, iso = s
            else:
                el, iso = s, None

            gammas.append(gyromagnetic_ratio(el, iso))
            Qs.append(quadrupole_moment(el, iso))
            Is.append(spin(el, iso))

            opdict = {a: SpinOperator.from_axes(Is[-1], a) for a in "xyz+-0"}

            operators.append(opdict)

        self._spins = list(spins)
        self._gammas = np.array(gammas)
        self._Qs = np.array(Qs)
        self._Is = np.array(Is)
        self._dim = tuple((2 * self._Is + 1).astype(int))

        self._operators = operators

        self._terms = []
        self._dissip_terms = []

        self._celio_k = celio_k

        snames = [
            "{1}{0}".format(*s) if (type(s) == tuple) else str(s) for s in self._spins
        ]
        logging.info("Created spin system with spins:")
        logging.info("\t\t{0}".format(" ".join(snames)))

    @property
    def spins(self):
        return list(self._spins)

    @property
    def celio_k(self):
        return self._celio_k

    @property
    def gammas(self):
        return self._gammas.copy()

    @property
    def Qs(self):
        return self._Qs.copy()

    @property
    def Is(self):
        return self._Is.copy()

    @property
    def dimension(self):
        return self._dim

    @property
    def is_dissipative(self):
        return (np.array(self._dissip_terms) != 0.0).any()

    def add_term(self, indices, tensor, label="Term"):
        """Add to the spin system a generic interaction term

        Add a term of the form T*S_i*S_j*S_k*..., where S_i is the vector of
        the three spin operators:

        [S_x, S_y, S_z]

        for spin of index i.

        Arguments:
            indices {[int]} -- Indices of spins appearing in the term
            tensor {ndarray} -- Tensor with n dimensions (n = len(indices)),
                                each of length 3, describing the interaction.

        Keyword Arguments:
            label {str} -- A label to name the term (default: {'Term'})

        Returns:
            term {InteractionTerm} -- The term just created

        Raises:
            ValueError -- Invalid index or vector
        """

        for i in indices:
            if i < 0 or i >= len(self._spins):
                raise ValueError("Invalid index i")

        tensor = np.array(tensor)

        term = InteractionTerm(self, indices, tensor, label=label)
        self._terms.append(term)

        return term

    def add_linear_term(self, i, vector, label="Single"):
        """Add to the spin system a term linear in one spin

        Add a term of the form v*S_i, where S_i is the vector of the three
        spin operators:

        [S_x, S_y, S_z]

        for spin of index i.

        Arguments:
            i {int} -- Index of the spin
            vector {ndarray} -- Vector v

        Keyword Arguments:
            label {str} -- A label to name the term (default: {'Single'})

        Returns:
            SingleTerm -- The term just created

        Raises:
            ValueError -- Invalid index or vector
        """

        if i < 0 or i >= len(self._spins):
            raise ValueError("Invalid index i")

        vector = np.array(vector)

        term = SingleTerm(self, i, vector, label=label)
        self._terms.append(term)

        return term

    def add_bilinear_term(self, i, j, matrix, label="Double"):
        """Add to the spin system a term bilinear in two spins

        Add a term of the form S_i*M*S_j, where S_i is the vector of the three
        spin operators:

        [S_x, S_y, S_z]

        for spin of index i, and same for S_j.

        Arguments:
            i {int} -- Index of first spin
            j {int} -- Index of second spin
            matrix {ndarray} -- Matrix M

        Keyword Arguments:
            label {str} -- A label to name the term (default: {'Double'})

        Returns:
            DoubleTerm -- The term just created

        Raises:
            ValueError -- Invalid index or vector
        """

        if i < 0 or i >= len(self._spins):
            raise ValueError("Invalid index i")

        if j < 0 or j >= len(self._spins):
            raise ValueError("Invalid index j")

        matrix = np.array(matrix)

        term = DoubleTerm(self, i, j, matrix, label=label)
        self._terms.append(term)

        return term

    def add_zeeman_term(self, i, B):
        """Add a zeeman term

        Add a single term coupling a given spin to a magnetic field

        Arguments:
            i {int} -- Index of the spin
            B {ndarray | number} -- Magnetic field vector, in Tesla. If just a
                                    scalar is assumed to be along z

        Returns:
            SingleTerm -- The term just created
        """

        if isinstance(B, Number):
            B = [0, 0, B]  # Treat it as along z by default

        B = np.array(B)

        logging.info("Adding Zeeman term to spin {0}".format(i + 1))

        return self.add_linear_term(i, B * self.gamma(i), "Zeeman")

    def add_dipolar_term(self, i, j, r):
        """Add a dipolar term

        Add a spin-spin dipolar coupling between two distinct spins. The
        coupling is calculated geometrically from the vector connecting them,
        in Angstrom.

        Arguments:
            i {int} -- Index of the first spin
            j {int} -- Index of the second spin
            r {ndarray} -- Vector connecting the two spins (in Angstrom)

        Returns:
            DoubleTerm -- The term just created

        Raises:
            ValueError -- Raised if i == j
        """

        if i == j:
            raise ValueError("Can not set up dipolar coupling with itself")

        r = np.array(r)

        g_i = self.gamma(i)
        g_j = self.gamma(j)

        rnorm = np.linalg.norm(r)
        D = -(np.eye(3) - 3.0 / rnorm**2.0 * r[:, None] * r[None, :])
        dij = -(cnst.mu_0 * cnst.hbar * (g_i * g_j * 1e6)) / (
            2 * (rnorm * 1e-10) ** 3
        )  # MHz
        D *= dij

        logging.info("Adding dipolar term to spins {0}-{1}".format(i + 1, j + 1))

        return self.add_bilinear_term(i, j, D, "Dipolar")

    def add_quadrupolar_term(self, i, EFG):
        """Add a quadrupolar term

        Add a quadrupolar term to a nucleus with I >= 1 from its Electric
        Field Gradient tensor.

        Arguments:
            i {int} -- Index of the spin
            EFG {ndarray} --  Electric Field Gradient tensor

        Returns:
            DoubleTerm -- The term just created
        """

        EFG = np.array(EFG)
        Q = self.Q(i)
        I = self.I(i)

        if I == 0.5:
            raise ValueError(
                "Can not set up quadrupolar coupling for spin 1/2 particle"
            )

        Qtens = EFG_2_MHZ * Q / (2 * I * (2 * I - 1)) * EFG

        logging.info("Adding quadrupolar term to spin {0}".format(i + 1))

        return self.add_bilinear_term(i, i, Qtens, "Quadrupolar")

    def remove_term(self, term):
        """Remove a term from the spin system

        Remove an interaction term from this spin system.

        Arguments:
            term {InteractionTerm} -- Term to remove

        Raises:
            ValueError -- The term is not contained in this system
        """

        self._terms.remove(term)

    def clear_terms(self):
        """Remove all terms

        Remove all interaction terms from this spin system.
        """

        terms = list(self._terms)
        for t in terms:
            self.remove_term(t)

    def add_dissipative_term(self, op, d=0.0):
        """Set a dissipation operator for the system.

        Set a dissipation operator for this system, representing its coupling
        (in MHz) with an external heat bath to include in the Lindbladian of
        the system.

        Arguments:
            op {SpinOperator} -- Operator for the dissipation term

        Keyword Arguments:
            d {number} -- Dissipation coupling in MHz (default: {0.0})
        """

        term = DissipationTerm(op, d)
        self._dissip_terms.append(term)

        return term

    def remove_dissipative_term(self, term):
        """Remove a dissipation term from the system.

        Remove a dissipation term from this spin system.

        Arguments:
            term {DissipationTerm} -- Term to remove

        Raises:
            ValueError -- The term is not contained in this system
        """

        self._dissip_terms.remove(term)

    def clear_dissipative_terms(self):
        """Remove all terms

        Remove all dissipative terms from this spin system.
        """

        dterms = list(self._dissip_terms)
        for t in dterms:
            self.remove_dissipative_term(t)

    def gamma(self, i):
        """Returns the gyromagnetic ratio of a given particle

        Arguments:
            i {int} -- Index of the particle

        Returns:
            float -- Gyromagnetic ratio in MHz/T
        """
        return self._gammas[i]

    def Q(self, i):
        """Returns the quadrupole moment of a given particle

        Arguments:
            i {int} -- Index of the particle

        Returns:
            float -- Quadrupole moment in Barn
        """
        return self._Qs[i]

    def I(self, i):
        """Returns the spin of a given particle

        Arguments:
            i {int} -- Index of the particle

        Returns:
            float -- Spin in units of hbar
        """

        return self._Is[i]

    def operator(self, terms={}, include_only_given=False):
        """Return an operator for this spin system
        Return a SpinOperator for this system containing the specified terms.
        Keyword Arguments:
            terms {dict} -- A dictionary of terms to include. The keys should
                            indices of particles and the values should be
                            symbols indicating one spin operator (either x, y,
                            z, +, - or 0). Wherever not specified, the identity
                            operator is applied (default: {{}})
            include_only_given -- When True only the requested terms will be included
                                  otherwise the result will include the kronecker
                                  product with identity matrices for the particles
                                  not present in the terms
        Returns:
            SpinOperator -- The requested operator
        """

        def _get_term(i):
            # Default to identity of the appropriate size if not specified
            # (and not requesting that we include only the explicitly
            # specified terms)
            ops = terms.get(i, None if include_only_given else "0")
            term_result = None
            if ops:
                # Compute matrix product of terms if a list is specified for
                # the index
                if isinstance(ops, list) and len(ops) > 0:
                    term_result = self._operators[i][ops[0]]
                    for op in ops[1:]:
                        term_result *= self._operators[i][op]
                else:
                    term_result = self._operators[i][terms.get(i, ops)]
            return term_result

        ops = []
        for i in range(len(self)):
            term = _get_term(i)
            if term:
                ops.append(term)

        M = ops[0]

        for i in range(1, len(ops)):
            M = M.kron(ops[i])

        return M

    def rotate(self, rotmat=np.eye(3)):

        # Trying to avoid pointlessly cloning the terms
        terms = self._terms
        self._terms = []

        # Make a clone
        rssys = self.clone()
        self._terms = terms

        # Edit the terms
        try:
            rssys._terms = [t.rotate(rotmat) for t in terms]
        except AttributeError:
            raise RuntimeError(
                "Can only rotate SpinSystems containing Single or Double terms"
            )

        return rssys

    @property
    def hamiltonian(self):
        H = None
        if not self._celio_k:
            if len(self._terms) == 0:
                n = np.prod(self.dimension)
                H = sparse.csr_matrix((n, n))
            else:
                H = np.sum([t.matrix for t in self._terms], axis=0)
            H = Hamiltonian(H, dim=self.dimension)
        else:
            H = CelioHamiltonian(self._terms, self._celio_k, self)

        return H

    @property
    def lindbladian(self):

        H = self.hamiltonian
        dops = [t.tuple for t in self._dissip_terms]
        L = Lindbladian.from_hamiltonian(H, dops)

        return L

    def __len__(self):
        return len(self._gammas)


class MuonSpinSystem(SpinSystem):
    def __init__(self, spins=["mu", "e"], celio_k=0):

        super(MuonSpinSystem, self).__init__(spins, celio_k)

        # Identify the muon index
        if self._spins.count("mu") != 1:
            raise ValueError(
                "Spins passed to MuonSpinSystem must contain exactly one muon"
            )

        self._mu_i = self._spins.index("mu")
        self._e_i = set([i for i, s in enumerate(self.spins) if s == "e"])

        # For convenience, store the operators for the muon
        self._mu_ops = [self.operator({self._mu_i: e}) for e in "xyz"]

    @property
    def muon_index(self):
        return self._mu_i

    @property
    def elec_indices(self):
        return self._e_i

    def add_hyperfine_term(self, i, A, j=None):
        """Add a hyperfine term

        Add a hyperfine term for a given spin, provided that an electron is
        present.

        Arguments:
            i {int} -- Index of the spin (must be different from electron)
            A {[type]} -- Hyperfine tensor (in MHz)
            j {int} -- Index of the electron spin. If not specified uses the
                       one that is present, if there is one (default: None)

        Returns:
            DoubleTerm -- The term just created

        Raises:
            ValueError -- Invalid index
        """

        elec_i = self.elec_indices

        if j is None:
            if len(elec_i) > 1:
                raise ValueError(
                    "Must specify an electron index in system with multiple "
                    "electrons"
                )
            else:
                j = list(elec_i)[0]
        else:
            if j not in elec_i:
                raise ValueError(
                    "Second index in hyperfine coupling must refer to an electron"
                )
        if i in elec_i:
            raise ValueError(
                "First index in hyperfine coupling must not refer to an electron"
            )

        logging.info("Adding hyperfine term to spins {0}-{1}".format(i + 1, j + 1))

        return self.add_bilinear_term(i, j, A, "Hyperfine")

    def muon_operator(self, v):
        """Get a muon operator
        Get a single operator for the muon, given a vector representing its
        direction. Uses precalculated operators for speed.
        Arguments:
            v {[float]} -- 3-dimensional vector representing the direction of
                           the desired operator
        Returns:
            mu_op {SpinOperator} -- Requested operator
        Raises:
            ValueError -- Invalid length of v
        """

        if len(v) != 3:
            raise ValueError("Vector passed to muon_operator must be three dimensional")

        op = [x * self._mu_ops[i] for i, x in enumerate(v)]
        op = sum(op[1:], op[0])

        return op

    def sigma_mu(self, v):
        """Obtain sigma_mu - a linear combination of Pauli spin operators in
        a given direction for the muon. Unlike the 'muon_operator' above it
        does not include a factor of spin and will be a 2x2 matrix instead of
        one as large as the total combined system.

        Arguments:
            v {[float]} -- 3-dimensional vector representing the direction of
                           the desired operator
        Returns:
            sigma_mu {ndarray} -- A linear combination of Pauli spin matrices
                                  in the direction of v
        Raises:
            ValueError -- Invalid length of v
        """

        if len(v) != 3:
            raise ValueError("Vector passed to muon_operator must be three dimensional")

        # Spin matrix in direction of the muon
        mu_ops = [sigmax().data, sigmay().data, sigmaz().data]
        sigma_mu = np.sum([x * mu_ops[i] for i, x in enumerate(v)])

        return sigma_mu



"""input.py

Class to read in input files for the muspinsim script
"""

import re
from io import StringIO

import numpy as np
from collections import namedtuple

from muspinsim.input.keyword import (
    InputKeywords,
    MuSpinEvaluateKeyword,
    MuSpinCouplingKeyword,
)

from muspinsim.input.larkeval import LarkExpressionError


class MuSpinInputError(Exception):
    pass


MuSpinInputValue = namedtuple("MuSpinInputValue", ["name", "args", "value"])

# Experiment defaults as .in files
_exp_defaults = {
    "alc": """
polarization
    longitudinal
y_axis
    integral
x_axis
    field
""",
    "zero_field": """
field
    0.0
polarization
    transverse
x_axis
    time
y_axis
    asymmetry
""",
}


def write_error(keyword, block_line_num, err):
    return (
        "Error occurred when parsing keyword '{0}' "
        "(block starting at line {1}):\n{2}".format(keyword, block_line_num, str(err))
    )


def _make_blocks(fs):
    lines = fs.readlines()

    # Split lines in blocks
    raw_blocks = {}
    curr_block = None
    block_line_nums = {}

    indre = re.compile("(\\s+)[^\\s]")
    indent = None

    for i, l in enumerate(lines):

        # Remove any comments
        l = l.split("#", 1)[0]

        if l.strip() == "":
            continue  # It's a comment
        m = indre.match(l)
        if m:
            if indent is None:
                indent = m.groups()[0]
            if m.groups()[0] != indent:
                raise RuntimeError("Invalid indent in input file")
            else:
                try:
                    raw_blocks[curr_block].append(l.strip())
                except KeyError:
                    raise RuntimeError("Badly formatted input file")
        else:
            curr_block = l.strip()
            raw_blocks[curr_block] = []
            block_line_nums[curr_block] = i + 1
            indent = None  # Reset for each block

    return raw_blocks, block_line_nums


class MuSpinInput(object):
    def __init__(self, fs=None):
        """Read in an input file

        Read in an input file from an opened file stream

        Arguments:
            fs {TextIOBase} -- I/O stream (should be file, can be StringIO)
        """

        self._keywords = {}
        self._variables = {}
        self._fitting_info = {"fit": False, "data": None, "method": None, "rtol": None}

        if fs is not None:

            raw_blocks, block_line_nums = _make_blocks(fs)

            # if we find errors when parsing fitting variables, we post an error
            # so we don't propagate invalid variables when parsing keywords later
            failed_status, errors = self._load_fitting_kw(raw_blocks, block_line_nums)
            if failed_status:
                raise MuSpinInputError(
                    "Found {0} Error(s) whilst trying to parse fitting keywords: "
                    "\n\n{1}".format(len(errors), "\n\n".join(errors))
                )

            # Another special case: if the "experiment" keyword is present,
            # use it to set some defaults
            errors_found = []
            try:
                block = raw_blocks.pop("experiment")
                kw = InputKeywords["experiment"](block)
                exptype = kw.evaluate()[0]
                try:
                    exp_kw, _ = _make_blocks(StringIO(_exp_defaults[exptype[0]]))
                    raw_blocks.update(exp_kw)
                except KeyError:
                    err = (
                        "Invalid experiment type '{0}' defined, "
                        "possible types include {1}".format(
                            exptype[0], list(_exp_defaults.keys())
                        )
                    )
                    errors_found += [
                        write_error("experiment", block_line_nums["experiment"], err)
                    ]
            except RuntimeError as e:
                errors_found += [
                    write_error("experiment", block_line_nums["experiment"], str(e))
                ]
            except KeyError:
                pass

            # Now parse
            for header, block in raw_blocks.items():
                hsplit = header.split()
                name = hsplit[0]
                args = hsplit[1:]
                try:
                    try:
                        KWClass = InputKeywords[name]
                    except KeyError:
                        raise RuntimeError(
                            "Invalid keyword {0} found in input file".format(name)
                        )

                    if issubclass(KWClass, MuSpinEvaluateKeyword):
                        kw = KWClass(block, args=args, variables=self._variables)
                    else:
                        kw = KWClass(block, args=args)

                    kwid = kw.id

                    if name != kwid:
                        self._keywords[name] = self._keywords.get(name, {})
                        self._keywords[name][kwid] = kw
                    else:
                        self._keywords[name] = kw
                except (ValueError, LarkExpressionError, RuntimeError) as e:
                    errors_found += [write_error(name, block_line_nums[header], str(e))]

            if errors_found:
                raise MuSpinInputError(
                    "Found {0} Error(s) whilst trying to parse keywords: "
                    "\n\n{1}".format(len(errors_found), "\n\n".join(errors_found))
                )

    @property
    def variables(self):
        return {**self._variables}

    @property
    def fitting_info(self):
        return {**self._fitting_info}

    def evaluate(self, **variables):
        """Produce a full dictionary with a value for every input keyword,
        interpreted given the variable values that have been passed."""

        result = {"couplings": {}, "fitting_info": self.fitting_info}

        for name, KWClass in InputKeywords.items():

            if issubclass(KWClass, MuSpinCouplingKeyword):
                if name in self._keywords:
                    for kwid, kw in self._keywords[name].items():
                        val = MuSpinInputValue(
                            name, kw.arguments, kw.evaluate(**variables)
                        )
                        result["couplings"][kwid] = val
            else:
                # remove unnecessary keywords - stored in fitting_info
                if name in [
                    "fitting_data",
                    "fitting_tolerance",
                    "fitting_variables",
                    "fitting_method",
                ]:
                    pass
                elif name in self._keywords:
                    kw = self._keywords[name]
                    v = variables if issubclass(KWClass, MuSpinEvaluateKeyword) else {}
                    val = kw.evaluate(**v)

                    result[name] = MuSpinInputValue(name, kw.arguments, val)

                elif KWClass.default is not None:
                    kw = KWClass()
                    val = np.array(kw.evaluate())

                    result[name] = MuSpinInputValue(name, kw.arguments, val)

        return result

    def _load_fitting_kw(self, raw_blocks, block_line_nums):
        """Special case: handling of all the fitting related keywords and
        information."""
        errors_found = []
        try:
            block = raw_blocks.pop("fitting_variables")
            kw = InputKeywords["fitting_variables"](block)
            self._variables = {v.name: v for v in kw.evaluate()}
        except KeyError:
            pass
        except (RuntimeError, ValueError, LarkExpressionError) as e:
            errors_found += [
                write_error(
                    "fitting_variables", block_line_nums["fitting_variables"], str(e)
                )
            ]

        if errors_found:
            return 1, errors_found

        if len(self._variables) == 0:
            return 0, None

        self._fitting_info["fit"] = True

        try:
            block = raw_blocks.pop("fitting_data")
            kw = InputKeywords["fitting_data"](block)
            self._fitting_info["data"] = np.array(kw.evaluate())
        except KeyError:
            errors_found += [
                write_error(
                    "fitting_variables",
                    block_line_nums["fitting_variables"],
                    "Fitting variables defined without defining any data to fit",
                )
            ]
        except (RuntimeError, ValueError, LarkExpressionError, IOError) as e:
            errors_found += [
                write_error("fitting_data", block_line_nums["fitting_data"], str(e))
            ]

        try:
            block = raw_blocks.pop("fitting_tolerance", [])
            kw = InputKeywords["fitting_tolerance"](block)
            self._fitting_info["rtol"] = float(kw.evaluate()[0][0])
        except (RuntimeError, ValueError) as e:
            errors_found += [
                write_error(
                    "fitting_tolerance", block_line_nums["fitting_tolerance"], str(e)
                )
            ]

        try:
            block = raw_blocks.pop("fitting_method", [])
            kw = InputKeywords["fitting_method"](block)
            self._fitting_info["method"] = kw.evaluate()[0][0]
        except (RuntimeError, ValueError) as e:
            errors_found += [
                write_error("fitting_method", block_line_nums["fitting_method"], str(e))
            ]

        if errors_found:
            return 1, errors_found
        return 0, None



import unittest
import numpy as np
from scipy import sparse

from muspinsim import constants
from muspinsim.spinop import SpinOperator, SuperOperator
from muspinsim.spinsys import (
    SpinSystem,
    InteractionTerm,
    SingleTerm,
    DoubleTerm,
    MuonSpinSystem,
)


class TestSpinSystem(unittest.TestCase):
    def test_create(self):

        ssys = SpinSystem(["mu", ("H", 2)])

        # Test cloning
        ssys2 = ssys.clone()

        self.assertEqual(ssys.dimension, ssys2.dimension)

        # Check that the copy is deep
        ssys2._spins[0] = "C"
        self.assertEqual(ssys.spins, ["mu", ("H", 2)])

    def test_terms(self):

        ssys = SpinSystem(["mu", "e"])

        # A scalar term
        term = InteractionTerm(ssys, [], 2.0)
        self.assertTrue(
            np.allclose(term.operator.matrix.data, (ssys.operator({}) * 2).matrix.data)
        )

        # Linear term
        term = SingleTerm(ssys, 0, [0, 0, 1])
        self.assertTrue(
            np.allclose(term.operator.matrix.data, ssys.operator({0: "z"}).matrix.data)
        )

        # Test copy
        copy = term.clone()
        copy._tensor[0] = 1

        self.assertTrue(isinstance(copy, SingleTerm))
        self.assertFalse(np.all(copy.tensor == term.tensor))

        # Bilinear term
        term = DoubleTerm(ssys, 0, 1, np.diag([1, 1, 2]))
        self.assertTrue(
            np.allclose(
                term.operator.matrix.data,
                (
                    ssys.operator({0: "x", 1: "x"})
                    + ssys.operator({0: "y", 1: "y"})
                    + 2 * ssys.operator({0: "z", 1: "z"})
                ).matrix.data,
            )
        )

        # Rotation matrix
        R = np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]])
        rotterm = term.rotate(R)
        self.assertTrue(
            np.allclose(
                rotterm.operator.matrix.data,
                (
                    ssys.operator({0: "x", 1: "x"})
                    + 2 * ssys.operator({0: "y", 1: "y"})
                    + ssys.operator({0: "z", 1: "z"})
                ).matrix.data,
            )
        )

    def test_check(self):

        ssys = SpinSystem(["mu", "e"])

        self.assertEqual(ssys.gamma(0), constants.MU_GAMMA)
        self.assertEqual(ssys.gamma(1), constants.ELEC_GAMMA)

        self.assertEqual(len(ssys), 2)
        self.assertEqual(ssys.dimension, (2, 2))

    def test_operator(self):

        ssys = SpinSystem(["mu", "e"])

        self.assertTrue(
            np.allclose(
                ssys.operator({0: "x"}).matrix.toarray(),
                SpinOperator.from_axes([0.5, 0.5], "x0").matrix.toarray(),
            )
        )

        self.assertTrue(
            np.allclose(
                ssys.operator({0: "z", 1: "y"}).matrix.toarray(),
                SpinOperator.from_axes([0.5, 0.5], "zy").matrix.toarray(),
            )
        )

        self.assertEqual(ssys.dimension, (2, 2))

    def test_operator_include_only_given(self):

        ssys = SpinSystem(["mu", "e"])

        self.assertTrue(
            np.allclose(
                sparse.kron(
                    ssys.operator({0: "x"}, True).matrix, sparse.identity(2)
                ).toarray(),
                SpinOperator.from_axes([0.5, 0.5], "x0").matrix.toarray(),
            )
        )

        self.assertTrue(
            np.allclose(
                ssys.operator({0: "z", 1: "y"}, True).matrix.toarray(),
                SpinOperator.from_axes([0.5, 0.5], "zy").matrix.toarray(),
            )
        )

        self.assertTrue(
            np.allclose(
                ssys.operator({0: ["z", "y"]}, True).matrix.toarray(),
                (
                    SpinOperator.from_axes(0.5, "z") * SpinOperator.from_axes(0.5, "y")
                ).matrix.toarray(),
            )
        )

        self.assertEqual(ssys.dimension, (2, 2))

    def test_addterms(self):

        ssys = SpinSystem(["mu", "e"])

        t1 = ssys.add_linear_term(0, [1, 0, 1])

        self.assertEqual(t1.label, "Single")

        self.assertTrue(
            np.allclose(
                t1.operator.matrix.data,
                (ssys.operator({0: "x"}) + ssys.operator({0: "z"})).matrix.data,
            )
        )

        t2 = ssys.add_bilinear_term(0, 1, np.eye(3))

        self.assertEqual(t2.label, "Double")
        self.assertTrue(
            np.allclose(
                t2.operator.matrix.data,
                (
                    ssys.operator({0: "x", 1: "x"})
                    + ssys.operator({0: "y", 1: "y"})
                    + ssys.operator({0: "z", 1: "z"})
                ).matrix.data,
            )
        )

        H = ssys.hamiltonian

        self.assertTrue(
            np.all(
                np.isclose(
                    H.matrix.toarray(),
                    np.array(
                        [
                            [0.75, 0, 0.5, 0.0],
                            [0, 0.25, 0.5, 0.5],
                            [0.5, 0.5, -0.75, 0],
                            [0.0, 0.5, 0, -0.25],
                        ]
                    ),
                )
            )
        )

        ssys.add_dissipative_term(ssys.operator({0: "x"}), 1.0)
        self.assertTrue(ssys.is_dissipative)

        # Now test clearing them
        ssys.clear_terms()

        H = ssys.hamiltonian

        self.assertTrue(np.all(np.isclose(H.matrix.toarray(), np.zeros((4, 4)))))

    def test_addterms_celio(self):

        ssys = SpinSystem(["mu", "e"], celio_k=10)

        t1 = ssys.add_linear_term(0, [1, 0, 1])

        self.assertEqual(t1.label, "Single")

        self.assertTrue(
            np.allclose(
                t1.operator.matrix.data,
                (
                    ssys.operator({0: "x"}, True) + ssys.operator({0: "z"}, True)
                ).matrix.data,
            )
        )

        t2 = ssys.add_bilinear_term(0, 1, np.eye(3))

        self.assertEqual(t2.label, "Double")
        self.assertTrue(
            np.allclose(
                t2.operator.matrix.data,
                (
                    ssys.operator({0: "x", 1: "x"}, True)
                    + ssys.operator({0: "y", 1: "y"}, True)
                    + ssys.operator({0: "z", 1: "z"}, True)
                ).matrix.data,
            )
        )

        evol_op_contribs = ssys.hamiltonian._calc_trotter_evol_op_contribs(1, False)
        self.assertEqual(len(evol_op_contribs), 2)

        self.assertTrue(
            np.all(
                np.isclose(
                    np.product(evol_op_contribs).toarray(),
                    np.array(
                        [
                            [
                                0.84425599 - 0.44143085j,
                                -0.09276182 - 0.01469203j,
                                0.04521741 - 0.28549151j,
                                0.0 + 0.0j,
                            ],
                            [
                                0.0 + 0.0j,
                                0.89336999 - 0.15115734j,
                                -0.049114 - 0.29027351j,
                                -0.0475444 - 0.30018354j,
                            ],
                            [
                                -0.0475444 - 0.30018354j,
                                0.13640963 - 0.26088945j,
                                0.80293516 + 0.41982569j,
                                0.0 + 0.0j,
                            ],
                            [
                                0.0 + 0.0j,
                                0.04521741 - 0.28549151j,
                                -0.09276182 - 0.01469203j,
                                0.93934479 + 0.15893624j,
                            ],
                        ]
                    ),
                )
            )
        )

        # Check C++ variant
        evol_op_contribs = ssys.hamiltonian._calc_trotter_evol_op_contribs(1, True)
        self.assertEqual(len(evol_op_contribs), 2)

        self.assertTrue(
            np.all(
                np.isclose(
                    evol_op_contribs[0].matrix,
                    np.array(
                        [
                            [
                                9.02916796e-01 - 0.30392537j,
                                6.93889390e-18 - 0.30392537j,
                            ],
                            [
                                -6.93889390e-18 - 0.30392537j,
                                9.02916796e-01 + 0.30392537j,
                            ],
                        ]
                    ),
                )
            )
        )
        self.assertEqual(evol_op_contribs[0].other_dim, 2)
        self.assertTrue(
            np.all(
                np.isclose(
                    evol_op_contribs[0].indices,
                    np.array(
                        [0, 1, 2, 3],
                    ),
                )
            )
        )

        self.assertTrue(
            np.all(
                np.isclose(
                    evol_op_contribs[1].matrix,
                    np.array(
                        [
                            [0.98768834 - 0.15643447j, 0, 0, 0],
                            [0, 0.93934743 + 0.14877802j, 0.04834091 - 0.30521248j, 0],
                            [0, 0.04834091 - 0.30521248j, 0.93934743 + 0.14877802j, 0],
                            [0, 0, 0, 0.98768834 - 0.15643447j],
                        ]
                    ),
                )
            )
        )
        self.assertEqual(evol_op_contribs[1].other_dim, 1)
        self.assertTrue(
            np.all(
                np.isclose(
                    evol_op_contribs[1].indices,
                    np.array(
                        [0, 1, 2, 3],
                    ),
                )
            )
        )

        # Now test clearing them
        ssys.clear_terms()

        evol_op_contribs = ssys.hamiltonian._calc_trotter_evol_op_contribs(1, False)

        self.assertEqual(len(evol_op_contribs), 0)

        # C++ variant
        evol_op_contribs = ssys.hamiltonian._calc_trotter_evol_op_contribs(1, True)

        self.assertEqual(len(evol_op_contribs), 0)

    def test_lindbladian(self):

        ssys = SpinSystem(["mu"])
        ssys.add_linear_term(0, [0, 0, 1])

        L = ssys.lindbladian
        L0 = -1.0j * SuperOperator.commutator(ssys.operator({0: "z"}))

        self.assertTrue(np.all(np.isclose(L.matrix.toarray(), L0.matrix.toarray())))

        sx = ssys.operator({0: "x"})
        d = 2.0
        ssys.add_dissipative_term(sx, d)

        L1 = L0 + d * (
            SuperOperator.bracket(sx) - 0.5 * SuperOperator.anticommutator(sx * sx)
        )

        L = ssys.lindbladian
        self.assertTrue(np.all(np.isclose(L.matrix.toarray(), L1.matrix.toarray())))

    def test_lindbladian_celio(self):

        ssys = SpinSystem(["mu"], 10)
        ssys.add_linear_term(0, [0, 0, 1])

        with self.assertRaises(NotImplementedError):
            L = ssys.lindbladian


class TestMuonSpinSystem(unittest.TestCase):
    def test_terms(self):

        mSsys = MuonSpinSystem(["e", "mu"])

        self.assertEqual(mSsys.dimension, (2, 2))
        self.assertEqual(mSsys.elec_indices, {0})
        self.assertEqual(mSsys.muon_index, 1)

        gmu = constants.MU_GAMMA
        ge = constants.ELEC_GAMMA

        z0 = mSsys.add_zeeman_term(0, 1.0)
        z1 = mSsys.add_zeeman_term(1, 1.0)
        h01 = mSsys.add_hyperfine_term(1, np.eye(3) * 100)

        H = mSsys.hamiltonian.matrix.toarray()

        self.assertTrue(
            np.all(
                np.isclose(
                    np.diag(H),
                    [
                        0.5 * (ge + gmu) + 25,
                        0.5 * (ge - gmu) - 25,
                        0.5 * (-ge + gmu) - 25,
                        0.5 * (-ge - gmu) + 25,
                    ],
                )
            )
        )

        mSsys.remove_term(h01)

        H = mSsys.hamiltonian.matrix.toarray()

        self.assertTrue(
            np.all(
                np.isclose(
                    np.diag(H),
                    [
                        0.5 * (ge + gmu),
                        0.5 * (ge - gmu),
                        0.5 * (-ge + gmu),
                        0.5 * (-ge - gmu),
                    ],
                )
            )
        )

        self.assertEqual(z0.label, "Zeeman")
        self.assertEqual(h01.label, "Hyperfine")

        mSsys = MuonSpinSystem(["e", "e", "mu"])

        self.assertEqual(mSsys.elec_indices, {0, 1})
        with self.assertRaises(ValueError):
            mSsys.add_hyperfine_term(2, np.eye(3))  # Must specify electron

    def test_muon_operator(self):
        mSsys = MuonSpinSystem(["mu", "e"])

        muon_operator = mSsys.muon_operator([1, 1, 1])
        self.assertEqual(muon_operator.matrix.shape, (4, 4))
        self.assertIsNone(
            np.testing.assert_array_equal(
                muon_operator.matrix.toarray(),
                [
                    [0.5 + 0.0j, 0.0 + 0.0j, 0.5 - 0.5j, 0.0 + 0.0j],
                    [0.0 + 0.0j, 0.5 + 0.0j, 0.0 + 0.0j, 0.5 - 0.5j],
                    [0.5 + 0.5j, 0.0 + 0.0j, -0.5 + 0.0j, 0.0 + 0.0j],
                    [0.0 + 0.0j, 0.5 + 0.5j, 0.0 + 0.0j, -0.5 + 0.0j],
                ],
            )
        )

    def test_sigma_mu(self):
        mSsys = MuonSpinSystem(["mu", "e"])

        muon_operator = mSsys.sigma_mu([1, 1, 1])
        self.assertEqual(muon_operator.shape, (2, 2))
        self.assertIsNone(
            np.testing.assert_array_equal(
                muon_operator.toarray(),
                [[1 + 0j, 1 - 1j], [1 + 1j, -1 + 0j]],
            )
        )



#!/usr/bin/env python
#
import setuptools
from glob import glob
import sys
import sysconfig
import os
import numpy

from pybind11.setup_helpers import Pybind11Extension, build_ext, ParallelCompile
from Cython.Build import cythonize


# Returns a boolean value from the environment variable with the given key
# if not found returns false
def _get_environ_bool(key):
    value = os.environ.get(key)
    if value is None:
        return False
    return value.lower() not in {"0", "false", ""}


def _check_user_arguments(options):
    # OpenMP
    options["openmp"] = "--with-openmp" in sys.argv or _get_environ_bool(
        "MUSPINSIM_WITH_OPENMP"
    )
    if "--with-openmp" in sys.argv:
        sys.argv.remove("--with-openmp")

    return options


# Platform selection similar to QuTiP's
def _add_compile_options(options):
    options["compile_args"] = []
    options["link_args"] = []

    # MSYSTEM assigned on Windows when using mingw
    if sysconfig.get_platform().startswith("win") and os.environ.get("MSYSTEM") is None:
        # Windows
        options["compile_args"].extend(["/w", "/Ox"])

        if options["openmp"]:
            options["compile_args"].append("/openmp")
            options["link_args"].append("/openmp")
    else:
        # Linux/Mac/Windows MINGW
        options["compile_args"].extend(["-w", "-O3", "-funroll-loops"])

        if options["openmp"]:
            options["compile_args"].append("-fopenmp")
            options["link_args"].append("-fopenmp")

    # MacOS specific
    if sysconfig.get_platform().startswith("macos"):
        # Needed for OSX 10.14+
        options["compile_args"].append("-mmacosx-version-min=10.9")
        options["link_args"].append("-mmacosx-version-min=10.9")

    return options


# Returns the extensions for compilation
def _get_extensions(options):
    ext_modules = []

    # Setup pybind11 extension
    ext_modules.append(
        Pybind11Extension(
            "muspinsim.cpp",
            sorted(glob("muspinsim/cpp/*.cpp")),
            # Need to copy the arrays as this step actually modifies them
            # and this breaks the cython compilation as -fvisibility=hidden
            # is used on linux which prevents the cython bindings being seen
            extra_compile_args=options["compile_args"].copy(),
            extra_link_args=options["link_args"].copy(),
        ),
    )

    # Setup cython extension
    ext_modules.extend(
        cythonize(
            [
                setuptools.Extension(
                    "muspinsim.cython",
                    sorted(glob("muspinsim/cython/*.pyx")),
                    extra_compile_args=options["compile_args"],
                    extra_link_args=options["link_args"],
                    # This requires numpy to be installed while building, so
                    # it's needed in setup_requires
                    include_dirs=[numpy.get_include()],
                )
            ],
            compiler_directives={"language_level": "3str"},
            # The following is useful when developing - generates a report to
            # see how much python is still used
            # annotate=True,
        ),
    )

    return ext_modules


def setup():
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    version = {}
    with open("muspinsim/version.py") as fp:
        exec(fp.read(), version)

    options = {}
    options = _check_user_arguments(options)
    options = _add_compile_options(options)

    # Optional multithreaded build for pybind11
    ParallelCompile("NPY_NUM_BUILD_JOBS").install()

    ext_modules = _get_extensions(options)

    setuptools.setup(
        name="muspinsim",
        version=version["__version__"],
        author="Simone Sturniolo",
        author_email="simonesturniolo@gmail.com",
        description="Full quantum simulation of muon experiments",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/muon-spectroscopy-computational-project/muspinsim.git",
        packages=setuptools.find_packages(),
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            "Development Status :: 5 - Production/Stable",
            # Indicate who your project is intended for
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering :: Chemistry",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Scientific/Engineering :: Information Analysis",
        ],
        # Includes are needed for Cython compilation
        setup_requires=["numpy"],
        install_requires=[
            "ase",
            "numpy",
            "scipy",
            "soprano",
            "lark",
            "qutip",
            "pybind11",
            "Cython",
        ],
        extras_require={
            "docs": ["mkdocs", "pymdown-extensions"],
            "dev": ["flake8", "black>=22.3.0", "pytest", "pre-commit"],
        },
        entry_points={
            "console_scripts": [
                "muspinsim = muspinsim.__main__:main",
                "muspinsim.mpi = muspinsim.__main__:main_mpi",
            ]
        },
        ext_modules=ext_modules,
        # Build tool for pybind11 - searches for highest supported C++ standard
        cmdclass={"build_ext": build_ext},
        python_requires=">=3.8",
    )


if __name__ == "__main__":
    setup()
